# ***************************************************************************
# Copyright 2014-2017 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import nsx_logging

from cli.utils.error_code import ErrorCode as ec
from cli.utils.error_utils import render_error


lg = nsx_logging.getLogger("cli.commands.cli_command_provider")


class CliCommandProvider(object):
    """Base class for local_python providers."""
    def __init__(self, descriptor):
        # The descriptor factory instantiates a local_python provider
        # and passes the corresponding CliCommandDescriptor instance.
        self._descriptor = descriptor

    def __call__(self, cmd_list):
        """Provider implementation for corresponding CliCommandDescriptor."""
        # Sub-classes implement this method to perform the actual
        # CLI command implementation.
        msg = ("Unimplemented __call__ method for descriptor: %s" %
               self._descriptor.id)
        lg.error(msg, error_code=ec.OPERATION_NOT_SUPPORTED)
        assert False, msg
        render_error("MSG_CLI_INTERNAL_ERROR")

    def show_configuration(self):
        """Provider's show configuration output, if any."""
        # Sub-classes implement this method to declare its "show configuration"
        # command. Typically only update type commands which modify system
        # configuration need to implement this. Also, the corresponding
        # descriptor yaml needs to declare a show_configuration_property.
        return None

    def get_context(self):
        """Return information about last known context.

        Return:
            (context, context_data) tuple where context can be one of the MODE_
            values defined in utils.constants and context_data is additional
            information associated with a context. For example when the user
            enters the interface context via "interface eth0", the value
            returned for context_data is eth0. context_data can also be None
            if no data is associated with a context, e.g. enable, configure,
            etc.
        """
        return self._descriptor.service.get_context()

    def set_context_data(self, new_data):
        """Set the context data

        new_data:
            the new context data, which will be kept for cli provider

        Return:
            (context, context_data) tuple)
        """
        return self._descriptor.service.set_context_data(new_data)

    def get_requires_restart(self):
        """Gets the status if nsx-cli restart is required.

        Return:
            bool: (requires_restart)
        """
        return self._descriptor.service.get_requires_restart()

    def set_requires_restart(self):
        """Sets the status of nsx-cli to restart is required.
        """
        return self._descriptor.service.set_requires_restart()

    def get_modifier(self):
        """Return modifier string specified for current command.

        Return:
            Modifier string specified by user for the current command, or
            None, if the user did not specify a modifier.
        """
        return self._descriptor.service.get_modifier()

    def get_arg_value(self, arg_name, cmd_list):
        """Return first value entered for arg_name in user specified cmd_list.

        Args:
            arg_name: Name of argument as declared by CliArgument.name
            cmd_list: User specified command list as passed to
                CliCommandProvider.__call__.
        Return:
            Value for argument arg_name specified by user; else, None if
            invalid arg_name. If arg_name is specified multiple times for
            for a command, the first value is returned.
        """
        values = self.get_arg_values(arg_name, cmd_list)
        return values[0] if values else None

    def get_arg_values(self, arg_name, cmd_list):
        """Return all values entered for arg_name in user specified cmd_list.

        Args:
            arg_name: Name of argument as declared by CliArgument.name
            cmd_list: User specified command list as passed to
                CliCommandProvider.__call__.
        Return:
            List of values for argument arg_name specified by user; [] if
            invalid argument name.
        Note:
            Most CliCommands will only have one instance of an argument
            arg_name. If this is the case, use the get_arg_value method (which
            returns a single value as opposed to a list of values) instead.
        """
        indexes = self._descriptor.get_arg_indexes(cmd_list).get(arg_name)
        if not indexes:
            return []
        indexes.sort()
        cmd_list_len = len(cmd_list)
        if not self._descriptor.additional_last_args:
            return [cmd_list[i] for i in indexes if i < cmd_list_len]
        path_len = len(self._descriptor.raw_schema.get("path", ""))
        result = []
        for index in indexes:
            if index >= cmd_list_len:
                break
            elif path_len - 1 == index:
                result.append(" ".join(cmd_list[index:]))
            else:
                result.append(cmd_list[index])
        return result

    def get_arg_value_by_keyword(self, keyword, cmd_list):
        """Return argument value for corresponding keyword in user specified
           cmd_list.

        Args:
            keyword: Keyword that preceeds argument in specified cmd_list
            cmd_list: User specified command list as passed to
                CliCommandProvider.__call__.
        Return:
            Value for argument that follows specified keyword.
        Note:
            This method assumes the same keyword is not used more than once
            in a command. So far this is true of all CLI commands; this method
            needs to be refactored if this is no longer the case.
        """
        arg_indexes = self._descriptor.get_arg_indexes(cmd_list)
        arg_names = list(arg_indexes.keys())
        for arg_name in arg_names:
            # For each index that contains an argument, look back one position
            # to see if there's a keyword match, if match, return argument value
            indexes = arg_indexes.get(arg_name, [])
            for index in indexes:
                if index and (keyword == cmd_list[index - 1]):
                    # The command has specified the last word in the command is
                    # an argument that can be repeated. Check if the index is
                    # the last word position, if so, return the additional
                    # argument values.
                    if self._descriptor.additional_last_args:
                        path = self._descriptor.raw_schema.get("path", "")
                        if len(path) - 1 == index:
                            return " ".join(cmd_list[index:])
                    return cmd_list[index]
        return None
