# ***************************************************************************
# Copyright 2015-2019 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import nsx_logging
import re
import socket
try:
    import simplejson
except ImportError:
    import json as simplejson
import sys

from cli.commands.cli_command_provider import CliCommandProvider
from cli.commands.host_shared.aph_info_config import _read_appliance_info_xml
from cli.commands.lcp_common.nsx_proxy.client import get_nsxproxy_cli_output
from cli.utils.command_utils import invoke_command
from cli.utils.constants import NODE_TYPE_CONTROLLER
from cli.utils.constants import NODE_TYPE_MANAGER
from cli.utils.constants import NODE_TYPE_HOST_ESX
from cli.utils.constants import NODE_TYPE_GLOBAL_MANAGER
from cli.utils.constants import NODE_TYPE_BAREMETAL_WINDOWS_SERVER
from cli.utils.constants import NODE_TYPE_PUBLIC_CLOUD_GATEWAY
from cli.utils.error_code import ErrorCode as ec
from cli.utils.error_utils import get_message
from cli.utils.node_utils import is_hypervisor
from cli.utils.node_utils import is_baremetal
from cli.utils.node_utils import get_node_type
from compat23 import compat23_httplib
from compat23.compat23_utils import compat23_JSONDecodeError


if is_hypervisor() or is_baremetal():
    from cli.utils.location_utils import FilePaths
else:
    NAPI_MPA_CONFIG_PATH = "/api/v1/node/mpa-config"
    NAPI_CONTROLLER_STATUS_PATH = "/api/v1/node/services/controller/status"
    from cli.utils.napiclient import request as napi_request


lg = nsx_logging.getLogger("cli.commands.common.get_managers")

CONNECTION_MSG_DICT = {
    "MSG_GET_MANAGERS_CONNECTED": "Connected",
    "MSG_GET_MANAGERS_STANDBY": "Standby"
}

def _get_connection_msg(msg):
    return CONNECTION_MSG_DICT[msg]


def _get_aphs():
    aph_list = _read_appliance_info_xml()
    return aph_list

def _check_connections(node_type):
    # Get the netstat output
    if node_type == NODE_TYPE_HOST_ESX:
        command = ["localcli", "network", "ip", "connection", "list"]
    elif node_type == NODE_TYPE_BAREMETAL_WINDOWS_SERVER:
        command = ["netstat", "-naop", "TCP"]
    else:
        # Root privileges are needed to display pids in netstat output
        command = ["sudo", "netstat", "-natp"]

    rc, out, err = invoke_command(command)

    if rc != 0:
        lg.error("Failed to get Manager connection status: %s", err,
                    error_code=ec.OPERATION_FAILED)
        return None
    else:
        return out


def _read_ccp_pid():
    status, json, _, _ = napi_request("GET", NAPI_CONTROLLER_STATUS_PATH)
    pids = json.get('pids', None)
    if isinstance(pids, list) and len(pids) > 0:
        return pids[0]
    return pids

def _get_managers_via_nsxproxy_cli(aph_list):
    """Retrieve APH connection status from nsxproxy cli

        In cases where NAT is involved, the manager IP may
        not be what we see in netstat output so we directly
        query TnProxy to check which connections are up.

    Return:
        A list of UUIDs for which a connection exists in TnProxy,
        or an empty list if cli could not be queried.
    """
    response = get_nsxproxy_cli_output("get/aph-conn-status")
    if response is None:
        return []

    # [{UUID, STATUS}]
    aph_status = simplejson.loads(response)['aph connection status']

    connected_aphs = []
    for aph in aph_status:
        uuid = aph['UUID']
        if aph['STATUS'] == 'CONNECTED':
            connected_aphs.append(uuid)
    return connected_aphs


def _get_managers_rpc(node_type, out, CliCommandProvider=None):
    if CliCommandProvider is not None:
        func_get_message = get_message
    else:
        func_get_message = _get_connection_msg

    aph_list = _get_aphs()
    if aph_list is None or out is None:
        return {}

    managers = {}

    if node_type == NODE_TYPE_PUBLIC_CLOUD_GATEWAY:
        # On PCG prefer querying nsxproxy CLI as netstat output
        # might be unreliable
        # This also immediately takes care of fqdn resolution
        connected_mgrs = set(_get_managers_via_nsxproxy_cli(aph_list))
        for aph in aph_list:
            if aph.uuid in connected_mgrs:
                status = func_get_message("MSG_GET_MANAGERS_CONNECTED")
            else:
                status = func_get_message("MSG_GET_MANAGERS_STANDBY")
            key = aph.fqdn if aph.fqdn is not None else aph.ip
            port = 443
            managers[key] = (None, port, status + " (NSX_RPC)", aph.is_master)
        return managers

    for aph in aph_list:
        ip_address = aph.ip
        port = aph.port
        if aph.fqdn is not None:
            try:
                ip_address = socket.gethostbyname(aph.fqdn)
            except:
                lg.error("Failed to resolve manager fqdn: %s", aph.fqdn,
                         error_code=ec.INTERNAL_ERROR)
                managers[aph.fqdn] = (None, None, "Unable to resolve fqdn", aph.is_master)
                continue

        connected = False
        conn_port = 0

        pattern = "{0}:{1}\s+ESTABLISHED".format(ip_address, port)
        if len(re.findall(pattern, out)) >= 1:
            connected = True
            conn_port = port
        # If no connection found, check if there is an inbound connection
        # to this host's port 443 (outbound connection from manager).
        else:
            pattern = "{0}:{1}\s+{2}:{3}\s+ESTABLISHED".format(
                            ip_address, "[0-9]+", "[0-9\.]+", 443)
            if len(re.findall(pattern, out)) >= 1:
                connected = True
                conn_port = 443

        # If neither of the above, we are not connected.
        if connected:
            status = func_get_message("MSG_GET_MANAGERS_CONNECTED")
        else:
            status = func_get_message("MSG_GET_MANAGERS_STANDBY")

        status_with_channel = status + " (NSX-RPC)"
        # TODO Certificate thumbprint?
        managers[ip_address] = (None, conn_port, status_with_channel, aph.is_master)
    return managers


def get_managers(CliCommandProvider=None):
    node_type = get_node_type()
    out = _check_connections(node_type)
    return _get_managers_rpc(node_type, out, CliCommandProvider)

class GetNodeManagers(CliCommandProvider):
    def __call__(self, _):
        mgrs = get_managers(CliCommandProvider)
        if mgrs is None:
            return "MSG_UNABLE_RETRIEVE_MANAGER_CONFIG"
        return "OK\n\n%s" % simplejson.dumps(mgrs)

if __name__ == '__main__':
    mgrs = get_managers()
    if mgrs is None:
        sys.stdout.write(
            "Unable to retrieve the list of configured managers\n")
    else:
        sys.stdout.write("%s" % simplejson.dumps(mgrs))
