# **********************************************************************
# Copyright 2020 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import subprocess
import PAM


_CHPASSWD_CMD = "/usr/sbin/chpasswd"


class PAMUtil(object):
    """
    A class that allows authenticating a user and changing a user's
    password via PAM.  Should be called in escalated mode so that we
    aren't prompted for the old password.
    """
    def __init__(self, username):
        self._username = username
        self._messages = []

    def authenticate(self, password):
        """
        Verify that the given password is correct.
        On success, returns a tuple (True, "")
        On failure, returns a tuple (False, "Error message")
        """
        def _conv(conv, msgs, appdata):
            del conv, appdata  # not used
            for msg in msgs:
                self._messages.append(msg)
            return [(password, 0)]

        result = False
        auth = PAM.pam()
        auth.start("nsx-node-api-managed-pam", self._username)
        auth.set_item(PAM.PAM_CONV, _conv)
        try:
            auth.authenticate()
            result = True
            msg = ""
        except PAM.error:
            result = False
            msg = "Authentication failure"
            # nsx-node-api-managed-pam doesn't insert delay. Sleep for a
            # random period in the range 1.0 - 5.0 seconds.
            gevent.sleep(random.choice(list(range(10, 51))) / 10.0)
        return result, msg

    def chpasswd_defunct(self, new_password):
        """
        Set user's password to new_password.
        On success, returns a tuple (True, "")
        On failure, returns a tuple (False, "Error message")

        FIXME: Due to python3-pam's failure on unified appliance (segfault),
        this method has been defunct. A new method (below) has replaced it,
        which directly calls command-line. In future this should be revisited.
        """
        def _conv(conv, msgs, appdata):
            del conv, appdata  # not used
            for msg in msgs:
                self._messages.append(msg)
            return [(new_password, 0)]

        result = False
        auth = PAM.pam()
        auth.start("chpasswd", self._username)
        auth.set_item(PAM.PAM_CONV, _conv)
        try:
            auth.chauthtok()
            result = True
            msg = ""
        except PAM.error:
            result = False
            # Return the last message sent to the conv method
            if len(self._messages) > 0:
                msg, _ = self._messages[-1]
            else:
                msg = "Unknown failure"

        return result, msg

    def chpasswd(self, new_password):
        cmd = [_CHPASSWD_CMD]
        p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stderr=subprocess.PIPE)
        _input = '%s:%s' % (self._username, new_password)
        _, err = p.communicate(input=_input.encode('utf-8'))
        err = err.strip()
        # The error output chpasswd emits can repeat the same error description
        # in mulitple lines, so we return the first line only.
        err = err.decode().split("\n")[0]
        success = (len(err) == 0)
        return success, err


def _authenticate(username, password):
    # Authenticate given user
    result, message = PAMUtil(username).authenticate(password)
    if not result:
        return 1, message
    return 0, ""


def _set_password(username, password):
    # Set specified username's password.
    result, message = PAMUtil(username).chpasswd(password)
    if not result:
        # #1734076: PAM reports "BAD PASSWORD: is too simple" if the new
        # password is too short, update message to indicate it may also
        # be too short
        if message == "BAD PASSWORD: is too simple":
            message = "BAD PASSWORD: is too simple or too short"
        return 1, message
    return 0, ""
