# ***************************************************************************
# Copyright 2019-2021 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
"""This utility script will read trap.json via stdin and then
write into .trp file"""
from datetime import datetime
import json
import os
import socket
import sys
import time
sys.path.append("/opt/vmware/nsx-common/python")
import nsx_logging
import traceback

# The reload module is not built-in for python 3
if sys.version_info.major == 3:
    from importlib import reload
# Reload the socket module to avoid using gevent socket
reload(socket)

_ADDRESS = "/var/run/snmp.ctl"

_COMPANY_OID = "6876"
_GENERIC_TRAP_TYPE = "6"
_TMP_FILENAME_EXTENSION = ".tmp"
_TRP_FILENAME_EXTENSION = ".trp"
_PATH_TRP = "/var/spool/snmp/"
_PATH_NODE_UUID_IN_MANAGER = "/common/configs/serial_number"
_PATH_NODE_UUID_IN_EDGE = "/config/vmware/node-uuid"
_PATH_NODE_UUID_IN_HOST = "/etc/vmware/nsx-mpa/mpaconfig.json"
_PATH_NODE_TYPE = "/etc/nsx_issue"
_PATH_VARBINDS_JSON = "/opt/vmware/nsx-monitoring/python/nsx_monitoring/varbinds.json"
_DOT = "."
_SEMICOLON = ";"
_PIPE = "|"
_INT_TYPE = "2"
_STR_TYPE = "4"
_DATE_TYPE = "100"
_NEWLINE = "\n"
_TIMESTAMP_OID = "1.3.6.1.4.1.6876.120.1.1.1"
_FEATURE_ID_OID = "1.3.6.1.4.1.6876.120.1.1.2"
_EVENT_TYPE_OID = "1.3.6.1.4.1.6876.120.1.1.3"
_EVENT_SEVERITY_OID = "1.3.6.1.4.1.6876.120.1.1.4"
_NODE_ID_OID = "1.3.6.1.4.1.6876.120.1.1.5"
_NODE_TYPE_OID = "1.3.6.1.4.1.6876.120.1.1.6"
_ENTITY_ID_OID = "1.3.6.1.4.1.6876.120.1.1.21"
_RETRY_COUNT = 3

_CVN_HV = "cvn-hv-"
_CVN_HV_LEN = len(_CVN_HV)
_NODE_TYPE = "node-type: "
_NODE_TYPE_LEN = len(_NODE_TYPE)
_NEW_LINE_IN_OCT = "#012"
_NSX_MANAGER = "nsx-manager"
_NSX_EDGE = "nsx-edge"

"""severity IDs are specified in monitoring/events/src/conf/VMWARE-NSX-MIB.mib."""
_SEVERITY_STR_TO_ID_MAP = {
    "emergency": 0,
    "alert": 1,
    "critical": 2,
    "error": 3,
    "warning": 4,
    "notice": 5,
    "info": 6,
    "debug": 7
}

"""node type IDs are specified in monitoring/events/src/conf/VMWARE-NSX-MIB.mib."""
_NODE_TYPE_STR_TO_ID_MAP = {
    "nsx-manager": 0,
    "global-manager": 0,
    "nsx-edge": 1,
    "nsx-esx": 2,
    "nsx-esxi": 2,
    "nsx-sleskvm": 3,
    "nsx-rhelkvm": 3,
    "nsx-kvm": 3,
    "nsx-public-cloud-gateway": 4,
    "nsx-intelligence": 5,
    "nsx-autonomous-edge": 6
}

lg = nsx_logging.getLogger(__name__)


_VARBINDS_DICT = {}
is_read_trap_json_failed = False


def read_trap_json(params_list):
    trp_list = []
    if params_list is None:
        try:
            params_list = json.load(sys.stdin)
        except Exception:
            lg.error("JSON load failure: %s", traceback.format_exc())
            return False, trp_list
    for params in params_list["events"]:
        status, file_name, trp_str = process_event(params)
        if status is True:
            trp_list.append((file_name,trp_str))
        else:
            lg.error("error happens when processing with %s", params)
            global is_read_trap_json_failed
            is_read_trap_json_failed = True
        #sleep 1 millisecond here to avoid duplicate trp file name
        time.sleep(.001)
    return True, trp_list


def process_event(params):
    trap_oid = params.get("trap_oid")
    if not trap_oid:
        lg.error('trap_oid is not in the json file or trap_oid field is empty for event %s', params)
        return False, "", ""
    enterprise_oid, specific_oid = get_enterprise_oid_and_specific_oid(trap_oid)
    time_in_ms = int(time.time() * 1000)

    file_name = str(time_in_ms) + "_" + _GENERIC_TRAP_TYPE + "_" + specific_oid +"_"+_COMPANY_OID

    feature_id = params.get("feature_id")
    if not feature_id:
        lg.error('feature_id is not in the json file or feature_id field is empty for event %s', params)
        return False, "", ""

    event_type = params.get("event_type")
    if not event_type:
        lg.error('event_type is not in the json file or event_type field is empty for event %s', params)
        return False, "", ""

    entity_id = params.get("entity_id")
    if not entity_id:
        lg.error('entity_id is not in the json file or entity_id field is empty for event %s', params)
        return False, "", ""

    event_severity = params.get("event_severity")
    if not event_severity:
        lg.error('event_severity is not in the json file or event_severity field is empty for event %s', params)
        return False, "", ""
    formatted_event_severity = convert_event_severity(event_severity)
    if formatted_event_severity == -1:
        lg.error('Invalid event severity %s for event %s', event_severity, params)
        return False, "", ""

    read_node_type_status, node_type = read_node_type()
    if read_node_type_status is False:
        lg.error('cannot read node_type from %s for event %s', _PATH_NODE_TYPE, params)
        return False, "", ""
    formatted_node_type = convert_node_type(node_type)
    if formatted_node_type == -1:
        lg.error('Invalid node_type %s for event %s', node_type, params)
        return False, "", ""

    read_node_id_status, node_id = read_node_id(formatted_node_type)
    if read_node_id_status is False:
        lg.error('cannot read node_id from %s or %s or %s for event %s',
                 _PATH_NODE_UUID_IN_MANAGER, _PATH_NODE_UUID_IN_EDGE, _PATH_NODE_UUID_IN_HOST, params)
        return False, "", ""

    trp_str = _DOT + enterprise_oid + _NEWLINE \
            + _GENERIC_TRAP_TYPE + _NEWLINE \
            + specific_oid + _NEWLINE \
            + _DOT + _TIMESTAMP_OID + _PIPE + _DATE_TYPE +_PIPE + get_formatted_time() + _NEWLINE\
            + _DOT + _FEATURE_ID_OID + _PIPE + _INT_TYPE + _PIPE + \
                str(feature_id) + _NEWLINE\
            + _DOT + _EVENT_TYPE_OID + _PIPE + _STR_TYPE + _PIPE + event_type + _NEWLINE\
            + _DOT + _EVENT_SEVERITY_OID + _PIPE + _INT_TYPE + _PIPE + \
                str(formatted_event_severity) + _NEWLINE\
            + _DOT + _NODE_ID_OID + _PIPE + _STR_TYPE + _PIPE + node_id + _NEWLINE\
            + _DOT + _NODE_TYPE_OID + _PIPE + _INT_TYPE + _PIPE + \
                str(formatted_node_type) + _NEWLINE\
            + _DOT + _ENTITY_ID_OID + _PIPE + _STR_TYPE + _PIPE + entity_id + _NEWLINE

    #read varbinds.json and add it to trp
    if params.get("context"):
        for cur_context_key in params.get("context"):
            if cur_context_key == "entity_id":
                continue
            cur_oid = cur_context_key.replace("-","_").upper() + "_OID"
            cur_context_value = params.get("context")[cur_context_key]
            if cur_oid in _VARBINDS_DICT:
                trp_str += _DOT + _VARBINDS_DICT[cur_oid] + _PIPE + _STR_TYPE + _PIPE + cur_context_value + _NEWLINE
            else:
                lg.error('{ %s, %s} is not allowed as a context', cur_context_key, cur_context_value)
    trp_str += _SEMICOLON + _NEWLINE
    return True, file_name, trp_str

def write_to_trp_file(trap_info, retry_count):
    for count in range(1, retry_count + 1):
        file_name = trap_info[0]
        trp_str = trap_info[1]
        try:
            with open(_PATH_TRP + file_name + _TMP_FILENAME_EXTENSION, "w") as f:
                f.write(trp_str)
            os.rename(_PATH_TRP + file_name + _TMP_FILENAME_EXTENSION, _PATH_TRP + file_name + _TRP_FILENAME_EXTENSION)
            return True
        except IOError as err:
            if count != retry_count:
                lg.error('Error creating trap file. Retry for another %d time. Error msg: %s',
                         retry_count - count, str(err))
            else:
                return False

def trigger_snmptrap_send():
    #create sockert
    sock = None
    try:
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(_ADDRESS)
        sock.send('N'.encode('utf-8'))
    except Exception:
        lg.exception('Error sending message with socket to %s: ', _ADDRESS)
        trp_files = os.listdir(_PATH_TRP)
        for trp_file in trp_files:
            if trp_file.endswith(".trp"):
                os.remove(os.path.join(_PATH_TRP, trp_file))
        return False
    finally:
        if sock:
            sock.close()
    return True

def get_enterprise_oid_and_specific_oid(trap_oid):
    return trap_oid.rsplit(".0.", 1)

def get_formatted_time():
    return datetime.utcnow().strftime('%Y%m%d%H%M%S.%f')[:-3]

def read_node_type():
    if not os.path.exists(_PATH_NODE_TYPE):
        lg.error('%s does not exists', _PATH_NODE_TYPE)
        return False, ""
    try:
        with open(_PATH_NODE_TYPE, "r") as f:
            lines = f.readlines()
        for line in lines:
            parts = line.split(":", 1)
            if parts[0].strip() == "node-type" and len(parts) > 1:
                node_type_info = parts[1].strip()
                break
        node_type_list = node_type_info.split(' ')
        for node_type in node_type_list:
            if node_type in _NODE_TYPE_STR_TO_ID_MAP:
                return True, node_type
        lg.error('node types %s are unexpected', node_type_list)
        return False, ""
    except IOError as err:
        lg.error('Error reading node type from %s. Error msg: %s',
                 _PATH_NODE_TYPE, str(err))
        return False, ""

def read_node_id(node_type_id):
    """ if it is manager or gm, check /common/configs/serial_number
        if it is edge, check /config/vmware/node-uuid
        if it is host, check /etc/vmware/nsx-mpa/mpaconfig.json """
    if node_type_id == _NODE_TYPE_STR_TO_ID_MAP["nsx-manager"]:
        if not os.path.exists(_PATH_NODE_UUID_IN_MANAGER):
            lg.error('%s is not exists', _PATH_NODE_UUID_IN_MANAGER)
            return False, ""
        try:
            with open(_PATH_NODE_UUID_IN_MANAGER) as f:
                lines = f.readlines()
            for line in lines:
                line = line.strip()
                parts = line.split("=", 1)
                if len(parts) == 2 and parts[0].lower() == "serial":
                    return True, parts[1].lower()
            lg.error('Cannot get manager node uuid from %s',
                     _PATH_NODE_UUID_IN_MANAGER)
            return False, ""
        except IOError as err:
            lg.error('Error reading manager node uuid from %s. Error msg: %s',
                     _PATH_NODE_UUID_IN_MANAGER, str(err))
            return False, ""
    elif node_type_id == _NODE_TYPE_STR_TO_ID_MAP["nsx-edge"]:
        if not os.path.exists(_PATH_NODE_UUID_IN_EDGE):
            lg.error('%s is not exists', _PATH_NODE_UUID_IN_EDGE)
            return False, ""
        try:
            with open(_PATH_NODE_UUID_IN_EDGE, "r") as f:
                node_uuid = f.read()
            if not node_uuid:
                lg.error('Cannot get edge node uuid from %s',
                     _PATH_NODE_UUID_IN_EDGE)
                return False, ""
            return True, node_uuid
        except IOError as err:
            lg.error('Error reading edge node uuid from %s. Error msg: %s',
                     _PATH_NODE_UUID_IN_EDGE, str(err))
            return False, ""
    else:
        if not os.path.exists(_PATH_NODE_UUID_IN_HOST):
            lg.error('%s is not exists', _PATH_NODE_UUID_IN_HOST)
            return False, ""
        try:
            with open(_PATH_NODE_UUID_IN_HOST, "r") as f:
                node_info_str = f.read()
            node_uuid = parse_node_uuid_in_host(node_info_str)
            if not node_uuid:
                lg.error('Cannot get host node uuid from %s',
                     _PATH_NODE_UUID_IN_HOST)
                return False, ""
            return True, node_uuid
        except IOError as err:
            lg.error('Error reading host node uuid from %s. Error msg: %s',
                     _PATH_NODE_UUID_IN_HOST, str(err))
            return False, ""

def parse_node_uuid_in_host(node_info_str):
    node_info_in_json = json.loads(node_info_str)
    account_name = node_info_in_json.get("AccountName")
    if account_name and account_name.startswith(_CVN_HV):
        node_uuid = account_name[_CVN_HV_LEN:]
    return node_uuid

def convert_event_severity(event_severity):
    return _SEVERITY_STR_TO_ID_MAP.get(event_severity.lower(), -1)

def convert_node_type(node_type):
    return _NODE_TYPE_STR_TO_ID_MAP.get(node_type, -1)

def read_varbinds():
    try:
        with open(_PATH_VARBINDS_JSON, "r") as f:
            varbinds_map = json.load(f)
        if not varbinds_map:
            return {}
        return varbinds_map
    except IOError as err:
        lg.error('Error reading varbinds definition from %s. Error msg: %s',
                 _PATH_VARBINDS_JSON, str(err))
        return {}

def exit_with_error(error_msg):
    lg.error(error_msg)
    sys.exit(error_msg)

def run(params_list=None):
    #check if varbinds.json is present
    global _VARBINDS_DICT
    if not _VARBINDS_DICT:
        if not os.path.exists(_PATH_VARBINDS_JSON):
            return _PATH_VARBINDS_JSON + " does not exist"
        _VARBINDS_DICT = read_varbinds()
    #check if /var/spool/snmp dir is present that is to say snmp is enabled
    #if SNMP agent is not enabled, exit with 0
    if not os.path.exists(_PATH_TRP):
        return None
    #if SNMP agent is enabled, send the trap and
    #-exit with 0 status if trap sent to target SNMP servers(if target
    # SNMP servers are configured),
    #-exit with non-zero status with the error log if trap not sent to
    # target SNMP servers(if target SNMP servers are configued)
    read_status, trp_list = read_trap_json(params_list)
    if not read_status:
        return "Failed to read trap json from standard input"
    for trp_info in trp_list:
        if not write_to_trp_file(trp_info, _RETRY_COUNT):
            lg.error("Error happened when writing to trp file")
    if not trigger_snmptrap_send():
        return "Failed to trigger snmptrap send"
    if is_read_trap_json_failed:
        return "Failed to send some of the SNMP traps"
    lg.info('SNMP traps are send out successfully')
    return None

if __name__ == "__main__":
    nsx_logging.basicConfig(syslog=True, subcomp="generate_snmptrp_file")
    lg.setLevel(nsx_logging.INFO)
    err = run()
    if err is not None:
        exit_with_error(err)
    sys.exit(0)
