"""EventSource class defintion"""
# ***************************************************************************
# Copyright 2019-2022 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import collections
import copy
import json
import os
import re
import sys
import time
import traceback
import uuid
from subprocess import PIPE
from subprocess import Popen

import gevent
from gevent.hub import LoopExit
from gevent.lock import BoundedSemaphore

if "win" in sys.platform:
    monitoring_path = \
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    nsx_path = \
        os.path.dirname(os.path.dirname(os.path.dirname(monitoring_path)))
    python_utils_path = \
        os.path.join(nsx_path, "nsx-python-utils", "python")
    python_logging_path = \
        os.path.join(nsx_path, "nsx-python-logging", "python")
    sys.path.append(python_utils_path)
    sys.path.append(python_logging_path)
    _EVENTS_JSON_PATH = os.path.join(monitoring_path, "events.json")
    _NODE_TYPE_PATH = os.path.join(nsx_path, "nsx_issue")
    _CUSTOMIZATION_JSON_PATH = os.path.join(nsx_path, "nsx-monitoring",
                                            "event_customizations.json")
else:
    sys.path.append("/opt/vmware/nsx-common/python")
    sys.path.append("/opt/vmware/nsx-monitoring/python")
    _EVENTS_JSON_PATH = \
        "/opt/vmware/nsx-monitoring/python/nsx_monitoring/events.json"
    _NODE_TYPE_PATH = "/etc/nsx_issue"
    _GENERATE_SNMPTRP_FILE_SCRIPT = \
        "/opt/vmware/nsx-monitoring/python/nsx_monitoring/bin/" \
        "generate_snmptrp_file.py"
    _CUSTOMIZATION_JSON_PATH = \
        "/var/run/vmware/nsx-monitoring/event_customizations.json"

from nsx_utils.node_utils import get_node_type
from nsx_utils.constants import NODE_TYPE_AUTONOMOUS_EDGE
from nsx_utils.constants import NODE_TYPE_BAREMETAL_WINDOWS_SERVER
from nsx_utils.constants import NODE_TYPE_HOST_CENTOSKVM
from nsx_utils.constants import NODE_TYPE_HOST_SLESKVM
from nsx_utils.constants import NODE_TYPE_HOST_RHELKVM
from nsx_utils.constants import NODE_TYPE_HOST_UBUNTUKVM
from nsx_utils.constants import NODE_TYPE_HOST_ESX

_NODE_TYPE = get_node_type()

# Currently Windows host does not support SNMP traps
_GENERATE_SNMPTRP_FILE_PATH = \
    "/opt/vmware/nsx-opsagent/libexec/nsx-generate-snmptrp-file"

if _NODE_TYPE in [NODE_TYPE_HOST_CENTOSKVM, NODE_TYPE_HOST_RHELKVM,
                  NODE_TYPE_HOST_SLESKVM, NODE_TYPE_HOST_UBUNTUKVM]:
    sys.path.append("/opt/vmware/nsx-common/lib/python")
    if _NODE_TYPE == NODE_TYPE_HOST_RHELKVM:
        it = iter(sys.modules['google'].__path__) if 'google' in sys.modules else None
        if (it is not None) and (next(it) not in
                                 ['/opt/vmware/nsx-netopa/lib/python/google',
                                  '/opt/vmware/nsx-common/lib/python/google']):
            del sys.modules['google']
elif _NODE_TYPE == NODE_TYPE_HOST_ESX:
    sys.path.append("/usr/lib/vmware/nsx-common/lib/python")

import nsx_logging

lg = nsx_logging.getLogger(__name__)

_EMIT_LOG_EVENT_STATE_ON = "On"
_EMIT_LOG_EVENT_STATE_OFF = "Off"

_ZERO_UUID = uuid.UUID("00000000-0000-0000-0000-000000000000")

if _NODE_TYPE in [NODE_TYPE_HOST_ESX, NODE_TYPE_BAREMETAL_WINDOWS_SERVER]:
    _PROVIDER_SOCKET = "tcp://127.0.0.1:9100"
else:
    _PROVIDER_SOCKET = \
        "unix:///var/run/vmware/nsx-opsagent/alarms-provider-service.sock"

_SEVERITY_TO_LOG_LEVEL = {
    "debug": nsx_logging.DEBUG,
    "info": nsx_logging.INFO,
    "warning": nsx_logging.WARNING,
    "error": nsx_logging.ERROR,
    "critical": nsx_logging.CRITICAL
}


class EventSource(object):
    """
        Used by Feature Teams to manage Events and update Alarms on the MP,
        see http://url/jgyn for a description of the Alarms/Events Framework.
    """
    def __init__(self, feature_id, source_id):
        """EventSource constructor

            Arguments:
            - feature_id: Integer feature ID assigned to Feature Team feature.
            - source_id: Source ID of daemon from where the Event instance was
                generated.
        """
        self.service = EventSourceService(feature_id, source_id)

    def set_poll_callback(self, poll_callback, event_type_id_list):
        """Register callback(s) for features that have is_poll_based Events.
            - Only called if a feature registers is_poll_based event types.
            - A feature can have different callbacks for different event_types.
            - Call once for each poll_callback with applicable event_type_ids.
            - The last callback registered for a given event_type_id is used.

            Callback details:
            - The poll_callback should be declared as follows:

                def poll_callback(event_type_id_integer_values_list):

                - returns an event_type_status_map dictionary in the form of:
                    {
                        event_type_id_integer_value :
                            {
                               entity_id_python_uuid_value : status_tuple
                            }
                    }
                - and status_tuple is in the form of:
                    (status_boolean_value, context)

            Arguments:
            - poll_callback: Callback to invoke for is_poll_based event_types.
            - event_type_id_list: List of applicable event_type_ids.

            Returns:
            - False if unknown feature_id, event_type_id combination specified,
              or if specified poll_callback is None.
        """
        return self.service.set_poll_callback(poll_callback, event_type_id_list)

    def set_config_callback(self, config_callback):
        """Register callback to receive event_type configuration changes.
            - Last invocation is the callback that's used.

            Arguments:
            - config_callback: Callback to invoke on event_type config changes.

            Callback details:
            - The config_callback should be declared as follows:

                def config_callback(event_type_config_map):

                - where event_type_config_map is in the form of:
                    {
                        event_type_id_integer_value : config_tuple
                    }
                - and config_tuple is in the form of:
                    (is_disabled_boolean_value, threshold_integer_value)

                For example, EventSource can invoke config_callback like so:
                    config_callback({21 : (False, 90), 22 : (True, 0)}

            Returns:
            - False if an unknown feature_id is specified in the constructor.
        """
        return self.service.set_config_callback(config_callback)

    def start(self):
        """Start managing Events.
            - Performs validation to ensure:
              * The self.feature_id is valid.
              * All is_poll_based True event_type_ids have callbacks configured.
              * A config callback is configured.
            - Establishes the channels to push Events to MP, SNMP, syslog.
            - Starts event processing.

            Returns:
            - False if any of the above validations fail.
        """
        return self.service.start()

    def update_event_instances(self, event_type_status_map):
        """Update non-is_poll_based (i.e. interrupt based) event types status.
            - Only called if feature supports non-is_poll_based event types.
            - Non-is_poll_based event types have status False on start.
            - Features must invoke this method on entity status change.
            - A non-is_poll_based event type instance status is the same as the
              last sample value until this method is invoked to set the next
              value.
            - Invocations of this method are ignored if self.start() not called.

            Arguments:
            - returns an event_type_status_map dictionary in the form of:
                {
                    event_type_id_integer_value :
                        {
                           entity_id_python_uuid_value : status_tuple
                        }
                }
            - and status_tuple is in the form of:
                (status_boolean_value, context)
        """
        self.service.update_event_instances(event_type_status_map)

    def update_deleted_entities(self, event_type_id, entity_id_list):
        """Update this EventSource instance of deleted entities.
            - Calling this method allows EventSrouce to clean up itself.
            - Invocations of this method are ignored if self.start() not called.

            Arguments:
            - event_type_id: Integer event type ID of deleted entities.
            - entity_id_list: list of entity Python.uuid instances.

            Notes:
            - If a feature cannot guarantee that it can invoke this method when
              its entities are deleted, then it must only declare is_poll_based
              Events. The Alarms/Events Framework is not responsible for Alarms
              stuck in the non-RESOLVED status because a feature no longer
              invokes this method for deleted entities.
        """
        self.service.update_deleted_entities(event_type_id, entity_id_list)

    def cluster_event_leader(self, is_leader, event_type_id_list):
        """Update EventSource instance with cluster_event:true event_type_ids.
            - Calling this method allows EventSource to manage
              cluster_event:true event_type_ids.
            - Features call this method to indicate the current node is or is
              not the leader node for the specified list of cluster_event:true
              event_type_ids.
            - It is the responsibillty of the feature to:
                - select a leader amongst the nodes where the feature and
                  cluster_event:true event_type_ids are applicable.
                - invoke this method on leadership change.

            Arguments:
            - is_leader: True if feature is the leader node for event_type_ids
              specified in event_type_id_list. If not the leader, specify False.
            - event_type_id_list: List of applicable event_type_ids.
        """
        self.service.cluster_event_leader(is_leader, event_type_id_list)


class EventLock(object):
    """A gevent Semaphore context manager to manage LoopExit and ValueError
       exceptions that can be raised when invoking the acquire() and
       release() methods"""
    def __init__(self, event_source_service, gevent_semaphore):
        self.service = event_source_service
        self.lock = gevent_semaphore

    def __enter__(self):
        try:
            self.lock.acquire()
        except LoopExit:
            # PR 2842168 indicates that EventSourceService.lock.acquire() can
            # raise a LoopExit exception, which may mean acquire() was called
            # consecutively without a release(). So we just catch the exception
            # and log a message to enable triage then keep going.
            lg.exception(
                "lock.acquire() LoopExit exception for "
                "feature_id: %s, source_id: %s",
                self.service.feature_id, self.service.source_id)
        return self

    def __exit__(self, exc_type, exc_value, tb):
        try:
            self.lock.release()
        except ValueError:
            lg.exception(
                "lock.release() ValueError exception for "
                "feature_id: %s, source_id: %s",
                self.service.feature_id, self.service.source_id)
        return False


class EventSourceService(object):
    """
        Used by an EventSource instance to perform the management of Events
        and update Alarms on the MP.
    """
    def __init__(self, feature_id, source_id):
        self.feature_id = feature_id
        self.source_id = source_id
        self.start_invoked_successfully = False
        # self.config_callback identifies the feature daemon's config callback
        self.config_callback = None
        # self.event_type_poll_callback identifies the feature daemon's callback
        # for each is_poll_based Event, it is a dictionary where:
        # - key: event_type_id
        # - val: poll_callback
        self.event_type_poll_callback = {}
        # self.event_type_configs tracks all Event metadata extracted from
        # events.json that are applicable to the specified feature_id and
        # current node type, it is a dictionary where:
        # - key: event_type_id
        # - val: event_dict
        self.event_type_configs = None
        # self.event_type_config_change tracks Event config changes the
        # Client Library needs to know about, it is a dictionary where:
        # - key: event_type_id
        # - val: event_dict of config values
        self.event_type_config_change = {}
        # self.event_type_config_sample_change tracks Event config changes the
        # feature daemon needs to know about, it is a dictionary where:
        # - key: event_type_id
        # - val: val: tuple(is_disabled, threshold).
        self.event_type_config_sample_change = {}
        # self.event_type_statuses tracks the last reported Event status (True
        # or False) after considering hysteresis of a given entity_id, it is a
        # dictionary where:
        # - key: event_type_id
        # - val: dictionary of form { entity_id (UUID) : tuple(status, context) }
        self.event_type_statuses = {}
        # self.event_type_history tracks the history of the all samples for a
        # given entity_id for histeresis purposes, it is a dictionary where:
        # - key: event_type_id
        # - val: dictionary of form { entity_id (UUID) : EventTypeHistory inst }
        # Note: the size of the list_of_bools is equal to the sample size
        # of a given Event.
        self.event_type_history = {}
        # Interval for reporting RPC errors is 300s (5 minutes)
        self.rpc_err_interval = 300
        # self.event_reports keeps a queue of event reports. The event reporter
        # thread will clear the queue periodically and sent event reports to
        # the provider.
        self.event_reports = collections.deque([])
        self.event_reports_queue_size = 10000
        self.event_reports_queue_lock = BoundedSemaphore()
        self._rpc_client_interval = 5
        # Following intervals need to be multiples of _rpc_client_interval
        self._event_report_interval = 5
        self._customization_interval = 5
        # Initialize the last modified time for customizations file
        self._customization_mtime = None
        if os.path.exists(_CUSTOMIZATION_JSON_PATH):
            self._customization_mtime = os.path.getmtime(
                _CUSTOMIZATION_JSON_PATH)
        # Setup the self.lock Semaphore which synchronizes access to self data
        self.lock = BoundedSemaphore()
        # self.schedule_manager tracks when event type instance samples should
        # be taken next
        self.schedule_manager = ScheduleManager(self)
        self.node_type = self._get_node_type(_NODE_TYPE_PATH)
        # Flag to track whether source id is valid
        self._valid_source_id = True
        # Flag to track whether reconnecting to provider is needed
        self._rpc_client_reconnect = False
        # self._leader_event_type_ids tracks all event_type_ids the feature
        # daemon is the leader of
        self._leader_event_type_ids = set([])

    def set_poll_callback(self, poll_callback, event_type_id_list):
        """Register callback(s) for features that have is_poll_based Events."""
        if not poll_callback:
            lg.warning(
                "EventSource client for feature_id %s specified an invalid "
                "poll_callback: %s", self.feature_id, poll_callback)
            return False
        if self.event_type_configs is None:
            if not self._load_events_json(_EVENTS_JSON_PATH):
                return False
        result = self._set_event_type_poll_callback(
            poll_callback, event_type_id_list)
        return result

    def set_config_callback(self, config_callback):
        """Register callback to receive event_type configuration changes."""
        if not config_callback:
            lg.warning(
                "EventSource client for feature_id %s specified an invalid "
                "config_callback: %s", self.feature_id, config_callback)
            return False
        if self.event_type_configs is None:
            if not self._load_events_json(_EVENTS_JSON_PATH):
                return False
        with EventLock(self, self.lock):
            self.config_callback = config_callback
        return True

    def start(self):
        """Start managing Events."""
        with EventLock(self, self.lock):
            return self._start_inner()

    def _start_inner(self):
        # Make sure the current node type was determined
        if self.node_type is None:
            lg.warning(
                "EventSource client for feature_id %s invoked start but "
                "the current node's node_type could not be determined",
                self.feature_id)
            return False
        # Make sure a config_callback is registered
        if not self.config_callback:
            lg.warning(
                "EventSource client for feature_id %s invoked start without "
                "specifying a config_callback", self.feature_id)
            return False
        # Make sure a poll_callback is registered for all is_poll_based events
        missing_callbacks = [
            event_type_id for event_type_id in self.event_type_poll_callback
            if self.event_type_poll_callback[event_type_id] is None]
        if missing_callbacks:
            lg.info(
                "EventSource client for feature_id %s invoked start without "
                "registering a poll_callback for event_type_ids: %s. These "
                "event_type_ids will be ignored.", self.feature_id,
                ", ".join([str(item) for item in missing_callbacks]))
            for event_type_id in missing_callbacks:
                self._remove_event_type_id(event_type_id)
        # Initialize self.event_type_statuses, self.event_type_history and
        # with all possible event_type_ids
        for event_type_id in self.event_type_configs:
            self.event_type_statuses[event_type_id] = {}
            self.event_type_history[event_type_id] = {}
        # Force an event customizations check so that the feature daemon's
        # config_callback is invoked and the feature daemon is notified of
        # the current event is_disabled and threshold values.
        try:
            self._check_customizations(force_check=True)
        except Exception:
            lg.exception(
                "EventSource client for feature_id %s invoked"
                "_check_customizations with force_check = True"
                "raised an exception", self.feature_id)
            return False
        # Start RPC client thread
        if self.node_type != NODE_TYPE_AUTONOMOUS_EDGE:
            gevent.spawn(self._rpc_client_thread)
            # Sleep 1s for the rpc client thread to receive status from provider
            gevent.sleep(1)
            # Make sure source id is valid for connection to provider
            if not self._valid_source_id:
                lg.warning("EventSource client for feature_id %s has duplicate "
                           "source id %s with other clients", self.feature_id,
                           self.source_id)
                return False
        # Everything looks good, okay to start client library processing
        self.start_invoked_successfully = True
        lg.info(
            "EventSource.start for feature_id %s completed successfully",
            self.feature_id)
        # Start scheduling
        self.schedule_manager.start_scheduling()
        return True

    def _remove_event_type_id(self, event_type_id):
        """Remove an event_type_id from all class dictionaries"""
        self.event_type_poll_callback.pop(event_type_id, None)
        self.event_type_configs.pop(event_type_id, None)
        self.event_type_config_change.pop(event_type_id, None)
        self.event_type_config_sample_change.pop(event_type_id, None)
        self.event_type_statuses.pop(event_type_id, None)
        self.event_type_history.pop(event_type_id, None)

    def update_event_instances(self, event_type_status_map):
        """Update non-is_poll_based (i.e. interrupt based) event types status"""
        if not self.start_invoked_successfully or not event_type_status_map:
            return
        self.process_new_sample(event_type_status_map, is_poll_based=False)

    def update_deleted_entities(self, event_type_id, entity_id_list,
                                disabled=False):
        """Update this EventSource instance of deleted entities."""
        if not self.start_invoked_successfully or not entity_id_list:
            return
        with EventLock(self, self.lock):
            for entity_id in entity_id_list:
                if (event_type_id in self.event_type_statuses and
                        entity_id in self.event_type_statuses[event_type_id]):
                    # If last reported Event for an entity_id is True, emit
                    # False to indicate the previous false-to-true transition
                    # no longer applies
                    status_tuple = \
                        self.event_type_statuses[event_type_id][entity_id]
                    if status_tuple[0]:
                        self._emit(event_type_id, entity_id, False,
                                   status_tuple[1], disabled=disabled)
                    del self.event_type_statuses[event_type_id][entity_id]
                if (event_type_id in self.event_type_history and
                        entity_id in self.event_type_history[event_type_id]):
                    del self.event_type_history[event_type_id][entity_id]

    def cluster_event_leader(self, is_leader, event_type_id_list):
        if not event_type_id_list:
            return
        # It's fine for features to call this method before invoking the start
        # method, if this happens, it just means we don't need to disconnect
        # and reconnect to the provider
        with EventLock(self, self.lock):
            reconnect_needed = False
            if is_leader:
                for event_type_id in event_type_id_list:
                    if event_type_id not in self._leader_event_type_ids:
                        self._leader_event_type_ids.add(event_type_id)
                        # reconnect is only needed if start() has been invoked
                        reconnect_needed = self.start_invoked_successfully
                    # EventSource was previously informed it's the leader for
                    # the current event_type_id, so it's a NOP
            else:
                for event_type_id in event_type_id_list:
                    if event_type_id in self.event_type_statuses:
                        entity_ids = list(
                            self.event_type_statuses[event_type_id].keys())
                        for _id in entity_ids:
                            del self.event_type_statuses[event_type_id][_id]
                    if event_type_id in self.event_type_history:
                        entity_ids = list(
                            self.event_type_history[event_type_id].keys())
                        for _id in entity_ids:
                            del self.event_type_history[event_type_id][_id]
                    try:
                        self._leader_event_type_ids.remove(event_type_id)
                    except KeyError:
                        pass  # EAFP
                    # Note: only the current leader should emit alarm, log
                    # message and SNMP trap, so on leadership loss, don't
                    # emit anything.
            if reconnect_needed:
                self._rpc_client_reconnect = True

    def _get_node_type(self, node_type_path):
        try:
            lines = []
            with open(node_type_path) as node_file:
                lines = node_file.readlines()
            for line in lines:
                parts = line.split(":", 1)
                if len(parts) < 2 or parts[0] != "node-type":
                    continue
                node_type = parts[1].strip()
                if " " not in node_type:
                    return node_type
                node_types = node_type
                parts = node_types.split()
                if ("nsx-manager" in parts or "nsx-controller" in parts or
                        "nsx-policy-manager" in parts):
                    node_type = "nsx-manager"
                else:
                    node_type = parts[0]
                    lg.warning(
                        "EventSource client for feature_id %s encountered an "
                        "unknown node type '%s', using the following node type "
                        "instead: %s", self.feature_id, node_types, node_type)
                return node_type
            lg.warning(
                "EventSource client for feature_id %s failed to determine "
                "node type: %s", self.feature_id, lines)
        except Exception:
            lg.exception(
                "EventSource client for feature_id %s failed to read the node "
                "type file: %s", self.feature_id, node_type_path)
        return None

    def _load_events_json(self, events_json_path):
        with EventLock(self, self.lock):
            return self._load_events_json_inner(events_json_path)

    def _load_events_json_inner(self, events_json_path):
        if self.event_type_configs:
            return True
        # Load events.json file
        events = {}
        try:
            with open(events_json_path) as json_file:
                objs = json.load(json_file)
        except Exception:
            lg.exception(
                "EventSource client for feature_id %s failed to load events "
                "JSON: %s", self.feature_id, events_json_path)
            return False
        # Get metadata for all events with self.feature_id
        for event_id in objs:
            event_dict = objs[event_id]
            feature_id = event_dict.get("feature_id")
            # events.json includes all Event definitions, filter out Events that
            # don't match the specified feature_id and Events with node_types
            # values that don't match the current node's node types.
            if not feature_id or feature_id != self.feature_id:
                continue
            if self.node_type not in event_dict.get("node_types", []):
                continue
            event_type_id = event_dict.get("event_type_id")
            events[event_type_id] = event_dict
            # Populate self.event_type_poll_callback
            if event_dict.get("is_poll_based"):
                self.event_type_poll_callback[event_type_id] = None
            # Populate self.event_type_config_change which tracks Event
            # config changes the Client Library needs to know about and
            # self.event_type_config_sample_change which tracks Event
            # config changes the feature daemon needs to know about
            self.event_type_config_change[event_type_id] = {
                "is_disabled": event_dict.get("is_disabled"),
                "threshold": event_dict.get("threshold"),
                "sensitivity": event_dict.get("sensitivity"),
                "suppress_alarm": event_dict.get("suppress_alarm"),
                "suppress_snmp_trap": event_dict.get("suppress_snmp_trap"),
                "sampling_interval": event_dict.get("sampling_interval")}
            self.event_type_config_sample_change[event_type_id] = tuple(
                [event_dict.get("is_disabled"),
                 event_dict.get("threshold")])
        if len(events) < 1:
            lg.warning(
                "EventSource client for feature_id %s did not find any events "
                "defined for it in events JSON", self.feature_id)
            return False
        self.event_type_configs = events
        return True

    def _set_event_type_poll_callback(self, poll_callback, event_type_id_list):
        with EventLock(self, self.lock):
            if self.event_type_configs:
                invalid_event_type_ids = []
                for event_type_id in event_type_id_list:
                    # Warn and ignore if invalid event_type_id is specified
                    if event_type_id not in self.event_type_configs:
                        invalid_event_type_ids.append(event_type_id)
                        continue
                    self.event_type_poll_callback[event_type_id] = poll_callback
                if invalid_event_type_ids:
                    lg.warning(
                        "EventSource client for feature_id %s specified the "
                        "following invalid event_type_ids when registering a "
                        "poll_callback", self.feature_id,
                        ", ".join([str(item)
                                   for item in invalid_event_type_ids]))
            return False

    def process_interrupt_based_samples(self, event_type_id):
        """Process interrupt based samples for the specified event_type_id."""
        with EventLock(self, self.lock):
            history_entities = self.event_type_history.get(event_type_id)
            for entity_id in history_entities:
                status_entities = self.event_type_statuses.setdefault(
                    event_type_id, {})
                last_status_tuple = \
                    status_entities.setdefault(entity_id, (False, {}))
                last_status = last_status_tuple[0]
                context = last_status_tuple[1]
                history = history_entities[entity_id]
                new_status = history.add_new_sample()
                if new_status != last_status:
                    self._emit(event_type_id, entity_id, new_status, context)
                    status_entities[entity_id] = (new_status, context)

    def process_new_sample(self, event_type_status_map, is_poll_based=True):
        """Process new samples."""
        with EventLock(self, self.lock):
            self._process_new_sample_inner(event_type_status_map, is_poll_based)

    def _process_new_sample_inner(self, event_type_status_map,
                                  is_poll_based=True):
        if not isinstance(event_type_status_map, dict):
            lg.warning(
                "EventSource._process_new_sample for feature_id %s "
                "received an invalid event_type_status_map: %s",
                self.feature_id, str(event_type_status_map))
            return
        for event_type_id in event_type_status_map:
            # If feature specified an invalid event_type_id, ignore it
            if event_type_id not in self.event_type_configs:
                lg.warning(
                    "EventSource._process_poll_sample for feature_id %s "
                    "recevied an invalid event_type_id: %s", self.feature_id,
                    event_type_id)
                continue
            elif (self.event_type_configs[event_type_id].get(
                    "is_poll_based", False) and not is_poll_based):
                lg.warning(
                    "EventSource._process_new_sample for feature_id %s "
                    "recevied an non-is_poll_based event_type_id: %s",
                    self.feature_id, event_type_id)
                continue
            elif self.event_type_config_sample_change[event_type_id][0]:
                lg.warning(
                    "EventSource._process_new_sample for feature_id %s "
                    "recevied a disabled event_type_id: %s",
                    self.feature_id, event_type_id)
                continue
            for entity_id in event_type_status_map[event_type_id]:
                # If feature specified an invalid entity_id type (not UUID)
                # or instance status (not boolean), ignore it
                if not isinstance(entity_id, uuid.UUID):
                    lg.warning(
                        "EventSource._process_new_sample for feature_id %s "
                        "received an invalid entity_id of type: %s",
                        self.feature_id, type(entity_id))
                    continue
                sample_status_tuple = \
                    event_type_status_map[event_type_id][entity_id]
                if not isinstance(sample_status_tuple, tuple):
                    lg.warning(
                        "EventSource._process_new_sample for feature_id %s "
                        "received an invalid sample_status_tuple of type: %s",
                        self.feature_id, type(sample_status_tuple))
                    continue
                sample_status = sample_status_tuple[0]
                if not isinstance(sample_status, bool):
                    lg.warning(
                        "EventSource._process_new_sample for feature_id %s "
                        "received an invalid sample_status of type: %s",
                        self.feature_id, type(sample_status))
                    continue
                # Verify context is json serializable
                context = sample_status_tuple[1]
                if context is None:
                    context = {}
                try:
                    json.dumps(context)
                except Exception:
                    lg.warning(
                        "EventSource._process_new_sample for feature_id "
                        "%s received an invalid context of type: %s",
                        self.feature_id, type(context))
                    continue
                # Determine the current status of the event_type_id.entity_id
                # if it's already being tracked, if not, update internal data
                # structures with new event_type_id.entity_id entry
                status_entities = self.event_type_statuses.setdefault(
                    event_type_id, {})
                last_status_tuple = \
                    status_entities.setdefault(entity_id, (False, {}))
                last_status = last_status_tuple[0]
                history_entities = self.event_type_history.setdefault(
                    event_type_id, {})
                if entity_id not in history_entities:
                    history = EventTypeHistory(
                        is_poll_based,
                        self.event_type_configs[event_type_id]["sample_size"],
                        self.event_type_config_change[event_type_id]["sensitivity"])
                    history_entities[entity_id] = history
                else:
                    history = history_entities[entity_id]
                if is_poll_based:
                    # Add the latest status sample to the history and compute
                    # the latest status considering hysteresis and emit if
                    # there's a status change
                    new_status = history.add_new_sample(sample_status)
                    if new_status != last_status:
                        self._emit(event_type_id, entity_id, new_status,
                                   context)
                        # Update status map with new status and context
                        status_entities[entity_id] = (new_status, context)
                else:
                    # Update the last interrupt status so that the next sample
                    # will get the latest interrupt status
                    current_status = False
                    if entity_id in status_entities:
                        current_status = status_entities[entity_id][0]
                    status_entities[entity_id] = (current_status, context)
                    history.set_last_interrupt_status(sample_status)

    def _emit(self, event_type_id, entity_id, new_status, context,
              disabled=False, syslog=True, alarm=True, snmp=True):
        # Callers of this method should have already acquired self.lock
        configs = self.event_type_configs[event_type_id]
        config_change = self.event_type_config_change[event_type_id]
        feature_id = configs["feature_id"]
        feature_name = configs["feature_name"]
        event_type = configs["event_type"]
        severity = configs["severity"]
        disallow_snmp_trap = configs["disallow_snmp_trap"]
        suppress_snmp_trap = config_change["suppress_snmp_trap"]
        suppress_alarm = config_change["suppress_alarm"]
        desc = configs["description"].replace('\n', ' ')
        desc_clear = configs["description_on_clear"].replace('\n', ' ')
        context_copy = None
        if context:
            keys = re.findall(r"{(.+?)}", desc if new_status else desc_clear)
            context_keys = list(context.keys())
            missing_keys = list(set(keys) - set(context_keys))
            extra_keys = list(set(context_keys) - set(keys))
            if missing_keys:
                lg.warning("Provided context for feature_id %s event_type_id "
                           "%s is missing the following keys: %s",
                           self.feature_id, event_type_id, str(missing_keys))
            if extra_keys:
                lg.warning("Provided context for feature_id %s event_type_id "
                           "%s has the following extra keys: %s",
                           self.feature_id, event_type_id, str(extra_keys))
            # remove invalid keys
            context_copy = {k: context[k] for k in context if k in keys}
            # add missing keys to support partial formatting
            for key in missing_keys:
                context_copy[key] = "{%s}" % key
            if new_status:
                desc = desc.format(**context_copy)
            else:
                desc_clear = desc_clear.format(**context_copy)
        if syslog:
            # Emit to syslog
            if new_status:
                msg = desc
                state = _EMIT_LOG_EVENT_STATE_ON
            else:
                if not disabled:
                    msg = desc_clear
                else:
                    msg = "Event %s.%s is now disabled" % (feature_name,
                                                           event_type)
                state = _EMIT_LOG_EVENT_STATE_OFF
            lg.log(
                _SEVERITY_TO_LOG_LEVEL[severity], msg,
                event_feature_name=feature_name, event_type=event_type,
                event_sev=severity, event_state=state, entity_id=entity_id)
        if alarm:
            # Insert event report into the send queue to be sent to provider
            if self.node_type != NODE_TYPE_AUTONOMOUS_EDGE and not \
                    suppress_alarm:
                # Skip sending the event if entity id is zero
                if entity_id == _ZERO_UUID:
                    lg.info("Skip sending event %s.%s because the entity id is"
                            " zero", feature_name, event_type)
                else:
                    lg.debug("Sending report to provider for event %s.%s",
                             feature_name, event_type)
                    with EventLock(self, self.event_reports_queue_lock):
                        self.event_reports.append(self.generate_single_event_report(
                            entity_id, new_status, context_copy, event_type_id))
        if snmp:
            # Emit to SNMP receiver
            if self.node_type != NODE_TYPE_BAREMETAL_WINDOWS_SERVER and not \
                    disallow_snmp_trap and not suppress_snmp_trap and snmp:
                oid_name = "event_%s_snmp_oid" % str(new_status).lower()
                trap_dict = {
                    "events": [{
                        "feature_id": feature_id,
                        "event_type": event_type,
                        "event_severity": severity,
                        "trap_oid": self.event_type_configs[
                            event_type_id][oid_name],
                        "entity_id": str(entity_id)
                    }]
                }
                if context_copy:
                    trap_dict["events"][0]["context"] = context_copy
                trap_json = json.dumps(trap_dict)
                lg.debug("Emitting to SNMP trap for event %s.%s", feature_name,
                         event_type)
                try:
                    if _NODE_TYPE == NODE_TYPE_HOST_ESX:
                        # On ESX the setuid script is not installed. All
                        # processes run as root and so call the python function
                        # directly.
                        from nsx_monitoring.bin import generate_snmptrp_file
                        err = generate_snmptrp_file.run(trap_dict)
                        if err is not None:
                            lg.warning("Failed to generate SNMP trap file: %s",
                                       err)
                    else:
                        po = Popen([_GENERATE_SNMPTRP_FILE_PATH], stdin=PIPE,
                                   stdout=PIPE, stderr=PIPE)
                        out, err = po.communicate(input=trap_json.encode())
                        if po.returncode != 0:
                            lg.warning("Failed to generate SNMP trap file: %s",
                                       err)
                except Exception:
                    lg.warning("Failed to execute %s for reason: %s",
                               _GENERATE_SNMPTRP_FILE_PATH, traceback.format_exc())

    # Helper function to create event_instance_status
    def create_event_instance_status(self, entity_id, status, context):
        from vmware.nsx.core_types_pb2 import UuidMsg
        from vmware.nsx.monitoring.monitoring_event_report_pb2 import \
            MonitoringEventInstanceStatusMsg
        entity_id_uuid = UuidMsg()
        entity_id_uuid.left = entity_id.int >> 64
        entity_id_uuid.right = entity_id.int & ((1 << 64) - 1)
        event_instance_status = MonitoringEventInstanceStatusMsg()
        event_instance_status.entity_id.CopyFrom(entity_id_uuid)
        event_instance_status.status = status
        if context:
            event_instance_status.context = json.dumps(context)
        return event_instance_status

    # Helper function to create event_type_status
    def create_event_type_status(self, event_type_id,
                                 event_instance_status_list):
        from vmware.nsx.monitoring.monitoring_event_report_pb2 import \
            MonitoringEventTypeStatusMsg
        event_type_status = MonitoringEventTypeStatusMsg()
        event_type_status.event_type_id = event_type_id
        event_type_status.monitoring_event_instance_status.extend(
            event_instance_status_list)
        return event_type_status

    # Helper function to create event_feature_report
    def create_event_feature_report(self, event_type_status_list):
        from vmware.nsx.monitoring.monitoring_event_report_pb2 import \
            MonitoringEventFeatureReportMsg
        event_feature_report = MonitoringEventFeatureReportMsg()
        event_feature_report.feature_id = self.feature_id
        event_feature_report.source_id = self.source_id
        event_feature_report.monitoring_event_type_status.extend(
            event_type_status_list)
        return event_feature_report

    # Helper function to create event_report
    def create_event_report(self, event_feature_report_list):
        from vmware.nsx.monitoring.monitoring_event_report_pb2 import \
            MonitoringEventReportMsg
        event_report = MonitoringEventReportMsg()
        event_report.monitoring_event_feature_report.extend(
            event_feature_report_list)
        return event_report

    def generate_single_event_report(self, entity_id, new_status, context,
                                     event_type_id):
        event_instance_status = \
            self.create_event_instance_status(entity_id, new_status, context)
        event_type_status = \
            self.create_event_type_status(event_type_id, [event_instance_status])
        event_feature_report = \
            self.create_event_feature_report([event_type_status])
        event_report = \
            self.create_event_report([event_feature_report])
        return event_report

    def generate_multi_event_report(self, event_type_instances):
        event_type_status_list = []
        # The event_type_instances argument is a list of entries in the form of
        # {<event_type_id>: <list of <entity id>>}
        for event_type_instance in event_type_instances:
            event_type_id = event_type_instance.event_type_id
            entity_ids = event_type_instance.entity_ids
            event_instance_status_list = []
            for entity_id in entity_ids:
                status_tuple = \
                    self.event_type_statuses.get(event_type_id).get(entity_id)
                event_instance_status_list.append(
                    self.create_event_instance_status(entity_id,
                                                      status_tuple[0], status_tuple[1]))
            event_type_status_list.append(self.create_event_type_status(
                event_type_id, event_instance_status_list))
        event_feature_report = \
            self.create_event_feature_report(event_type_status_list)
        event_report = \
            self.create_event_report([event_feature_report])
        return event_report

    def generate_full_event_report(self):
        event_type_status_list = []
        for event_type_id, statuses in self.event_type_statuses.items():
            event_instance_status_list = []
            # Create list of event_instance_status
            for entity_id, status_tuple in statuses.items():
                if entity_id == _ZERO_UUID:
                    continue  # Don't emit to MP if node uuid is still zeros
                event_instance_status_list.append(
                    self.create_event_instance_status(entity_id,
                                                      status_tuple[0], status_tuple[1]))
            event_type_status_list.append(self.create_event_type_status(
                event_type_id, event_instance_status_list))
        event_feature_report = \
            self.create_event_feature_report(event_type_status_list)
        event_report = \
            self.create_event_report([event_feature_report])
        return event_report

    def get_event_type_schedule_details(self):
        """Get schedule details of all event types."""
        # Callers of this method should have already acquired self.lock
        results = []
        for event_type_id, event_dict in self.event_type_config_change.items():
            if not event_dict["is_disabled"]:
                results.append((
                    event_type_id, event_dict["sampling_interval"],
                    self.event_type_poll_callback.get(event_type_id)))
            else:
                lg.info("Skipping disabled feature_id %d event_type_id %s",
                        self.feature_id, event_type_id)
        return results

    def _str_to_bool(self, val):
        return val.lower() == "true" if isinstance(val, str) else val

    def _format_customizations_json(self, json_objs):
        for name, props in json_objs.items():
            props["is_disabled"] = \
                self._str_to_bool(str(props["is_disabled"]))
            props["suppress_alarm"] = \
                self._str_to_bool(str(props["suppress_alarm"]))
            props["suppress_snmp_trap"] = \
                self._str_to_bool(str(props["suppress_snmp_trap"]))
            props["sampling_interval"] = int(props["sampling_interval"])
            props["sensitivity"] = int(props["sensitivity"])
            props["threshold"] = int(props["threshold"])

    def _check_customizations(self, force_check=False):
        try:
            sample_configs_changed = False
            if os.path.exists(_CUSTOMIZATION_JSON_PATH):
                mtime = os.path.getmtime(_CUSTOMIZATION_JSON_PATH)
                if (force_check or self._customization_mtime is None or
                        mtime != self._customization_mtime):
                    lg.info("Found updates in event customizations json file, "
                            "checking changes for feature_id %d",
                            self.feature_id)
                    # Load customizations.json
                    with open(_CUSTOMIZATION_JSON_PATH) as json_file:
                        objs = json.load(json_file)
                        self._format_customizations_json(objs)
                    invalid_event_types = []
                    for key, props in objs.items():
                        feature_name = key.split(".")[0]
                        feature_match = True
                        for id, configs in self.event_type_configs.items():
                            if configs["feature_name"] != feature_name:
                                feature_match = False
                            break
                        if not feature_match:
                            continue
                        event_type = key.split(".")[1]
                        event_type_id = None
                        for id, configs in self.event_type_configs.items():
                            if configs["event_type"] == event_type:
                                event_type_id = id
                                break
                        if not event_type_id:
                            invalid_event_types.append(event_type)
                            continue
                        # Update event_type_config_change for changes
                        conf = self.event_type_config_change[event_type_id]
                        if conf != props:
                            self.event_type_config_change[event_type_id] = \
                                copy.deepcopy(props)
                            lg.info("Update config changes for feature id %d "
                                    "event type id %d to %s", self.feature_id,
                                    event_type_id, str(props))
                        if not props["is_disabled"]:
                            # If suppress_alarm transitions to False, send
                            # event report for the entity ids with True status
                            if conf["suppress_alarm"] and not \
                                    props["suppress_alarm"]:
                                lg.info("suppress_alarm is disabled, resending "
                                        "reports for active %s events", event_type)
                                for entity_id, status_tuple in \
                                        self.event_type_statuses[
                                            event_type_id].items():
                                    if status_tuple[0]:
                                        self._emit(event_type_id, entity_id,
                                                   True, status_tuple[1], syslog=False,
                                                   alarm=True, snmp=False)
                            # If suppress_snmp_trap transitions to False, send
                            # SNMP trap for the entity ids with True status
                            if conf["suppress_snmp_trap"] and not \
                                    props["suppress_snmp_trap"]:
                                lg.info("suppress_snmp_trap is disabled, "
                                        "resending SNMP traps for active %s events",
                                        event_type)
                                for entity_id, status_tuple in \
                                        self.event_type_statuses[
                                            event_type_id].items():
                                    if status_tuple[0]:
                                        self._emit(event_type_id, entity_id,
                                                   True, status_tuple[1], syslog=False,
                                                   alarm=False, snmp=True)
                        # Update sensitivity for all entity ids
                        if conf["sensitivity"] != props["sensitivity"]:
                            lg.info("Updating sensitivity for feature id %d, "
                                    "event type id %d", self.feature_id,
                                    event_type_id)
                            for entity_id in \
                                    self.event_type_history[event_type_id]:
                                self.event_type_history[event_type_id][
                                    entity_id].set_sensitivity(
                                    props["sensitivity"])
                        # Update sampling interval for all entity ids
                        if conf["sampling_interval"] != props[
                            "sampling_interval"]:
                            lg.info("Updating sampling interval for feature id "
                                    "%d, event type id %d", self.feature_id,
                                    event_type_id)
                            self.schedule_manager._update_schedule(
                                event_type_id, props["sampling_interval"],
                                self.event_type_poll_callback.get(
                                    event_type_id))
                        # Update event_type_config_sample_change for changes
                        sample_conf = self.event_type_config_sample_change[
                            event_type_id]
                        sample_props = (props.get("is_disabled"),
                                        props.get("threshold"))
                        if sample_conf != sample_props:
                            self.event_type_config_sample_change[
                                event_type_id] = sample_props
                            sample_configs_changed = True
                            lg.info("Update config sample changes for feature "
                                    "id %d event type id %d to %s",
                                    self.feature_id, event_type_id,
                                    str(sample_props))
                        # Add schedule for events newly enabled
                        if conf["is_disabled"] and not props["is_disabled"]:
                            lg.info("Adding schedule for feature_id %d "
                                    "event_type_id %s", self.feature_id,
                                    event_type_id)
                            self.schedule_manager._add_new_schedule(
                                event_type_id, props["sampling_interval"],
                                self.event_type_poll_callback.get(
                                    event_type_id))
                        if props["is_disabled"] and not conf["is_disabled"]:
                            lg.info("Disabling feature_id %d event_type_id "
                                    "%s", self.feature_id, event_type_id)
                            if self.event_type_statuses[event_type_id]:
                                # Invoking update_deleted_entities will
                                # emit T-to-F transisions and clean up
                                # the data structures
                                self.update_deleted_entities(event_type_id,
                                                             list(self.event_type_statuses[
                                                                      event_type_id].keys()), True)
                    if invalid_event_types:
                        lg.debug("Ignoring the following event types in event "
                                 "customizations for feature_id %s since they "
                                 "are not registered in the client library: %s",
                                 self.feature_id,
                                 ", ".join([str(item) for item in
                                            invalid_event_types]))
                    self._customization_mtime = mtime
            # If force_check is True or sample configs map changed,
            # notify feature daemons
            if force_check or sample_configs_changed:
                event_type_config_map = copy.deepcopy(
                    self.event_type_config_sample_change)
                self.config_callback(event_type_config_map)
        except Exception:
            lg.warning("Customization check for feature_id %d failed with "
                       "error reason: %s", self.feature_id,
                       traceback.format_exc())

    def send_event_reports(self, process_event_report):
        try:
            while self.event_reports:
                response = process_event_report(self.event_reports[0])
                if not response.result:
                    lg.warning("processEventReport returned False")
                    return False
                self.event_reports.popleft()
            return True
        except Exception:
            lg.warning("Event reporter for feature_id %d failed with error "
                       "reason: %s", self.feature_id, traceback.format_exc())
            return False

    def interval_lcm(self, intervals):
        sum = 0
        while (True):
            lcm_found = True
            sum += self._rpc_client_interval
            for interval in intervals:
                if sum % interval != 0:
                    lcm_found = False
                    break
            if lcm_found:
                break
        return sum

    def _rpc_client_thread(self):
        try:
            # Import RPC classes
            from vmware.nsx.rpc import NsxRpcClient
            from vmware.nsx.monitoring.monitoring_provider_service_pb2 import \
                MonitoringProviderService_Stub
            from vmware.nsx.monitoring.monitoring_provider_service_pb2 import \
                MonitoringSubClientSubscribeMsg
            from vmware.nsx.monitoring.monitoring_provider_service_pb2 import \
                MonitoringSubClientRequestMsg
        except ImportError:
            lg.warning("Failed to import nsx-monitoring package, skip starting"
                       " the RPC client greenlet")
            return
        # Initialize timers and threshold
        rpc_err_timer = 0
        interval_threshold = self.interval_lcm([
            self._event_report_interval,
            self._customization_interval
        ])
        lg.info("RPC client thread started")
        # Start RPC connection
        while True:
            if not self._valid_source_id:
                lg.info("Skip connecting to provider due to invalid source_id")
                break
            try:
                with NsxRpcClient(MonitoringProviderService_Stub,
                                  _PROVIDER_SOCKET) as nsx_rpc_client:
                    lg.info("Connected to provider at %s", _PROVIDER_SOCKET)
                    # Reset timer for each newly established connection
                    rpc_client_timer = 0
                    process_event_report = \
                        eval("nsx_rpc_client.ProcessEventReport")
                    try:
                        # Subscribe_async
                        callback = Callback(self.source_id,
                                            MonitoringSubClientRequestMsg,
                                            self, nsx_rpc_client)
                        client_context = nsx_rpc_client.CreateClientContext()
                        client_context.SetClientId(self.source_id)
                        subscribe_msg = MonitoringSubClientSubscribeMsg()
                        subscribe_msg.source_id = self.source_id
                        nsx_rpc_client.Subscribe_async(
                            subscribe_msg, callback, client_context)
                    except Exception:
                        lg.warning("Callback subscribe error: %s",
                                   traceback.format_exc())
                    # Once RPC connection established, keep it alive by staying
                    # inside the with block
                    while True:
                        if self._rpc_client_reconnect:
                            lg.warning("RPC connection is broken")
                            self._rpc_client_reconnect = False
                            break
                        # Reset timer when it exceeds least common multiple of
                        # all intervals so that it does not overflow
                        if rpc_client_timer > interval_threshold:
                            rpc_client_timer %= interval_threshold
                        if rpc_client_timer % self._customization_interval == 0:
                            self._check_customizations()
                        if rpc_client_timer % self._event_report_interval == 0:
                            with EventLock(self, self.event_reports_queue_lock):
                                if len(self.event_reports) >= \
                                        self.event_reports_queue_size:
                                    lg.info("Event report queue is full, "
                                            "clearing the queue and reconnecting to"
                                            " the provider to trigger full sync")
                                    self.event_reports.clear()
                                    # Exit inner loop to reconnect, which will
                                    # trigger a full sync request from provider
                                    break
                            if not self.send_event_reports(
                                    process_event_report):
                                # Exit inner loop to reconnect if sending report
                                # failed with any exceptions
                                lg.warning("Sending event report failed")
                                break
                        gevent.sleep(self._rpc_client_interval)
                        rpc_client_timer += self._rpc_client_interval
            except Exception as ex:
                if rpc_err_timer > self.rpc_err_interval:
                    rpc_err_timer %= self.rpc_err_interval
                # TODO: remove comments when the provider component is available
                if rpc_err_timer % self.rpc_err_interval == 0:
                    lg.warning("RPC client failed with error reason: %s",
                               traceback.format_exc())
            finally:
                # Retry connection in next wake-up
                gevent.sleep(self._rpc_client_interval)
                rpc_err_timer += self._rpc_client_interval


class Callback(object):
    _MSG_TYPE_NAMES = {
        0: "INVALID",
        1: "SUBSCRIBE_ACK",
        2: "FULL_SYNC_REQ",
        3: "SYNC_REQ"
    }

    def __init__(self, source_id, response_class, event_source_service,
                 nsx_rpc_client):
        self.source_id = source_id
        self.response_class = response_class
        self.service = event_source_service
        self.nsx_rpc_client = nsx_rpc_client

    def __call__(self, message, status):
        if message is not None:
            gevent.spawn(self._message_processor_thread, message)
        if status is not None:
            lg.info("Received status for %s: '%s'" % (self.source_id, status))
            if not status.is_success():
                if status.name == "ALREADY_EXISTS":
                    # Provider returns ALREADY_EXISTS status if source_id
                    # conflicts with other clients
                    self.service._valid_source_id = False
                if status.name == "COMMUNICATION_ERROR" and \
                        "LOCAL_CLOSE" in status.message:
                    # Ignore LOCAL_CLOSE status
                    return
                # If the streaming client receives a status that is not
                # SUCCESS or LOCAL_CLOSE, set the reconnect flag to retry
                # connection
                self.service._rpc_client_reconnect = True

    def _message_processor_thread(self, message):
        lg.info("Received %s message for %s" %
                (self._MSG_TYPE_NAMES.get(message.msg_type),
                 self.source_id))
        assert isinstance(message, self.response_class)
        if message.msg_type in [2, 3]:
            send_msg = eval("self.nsx_rpc_client.SendMsg")
            # FullSync request
            if message.msg_type == 2:
                self._do_full_sync(send_msg, message.correlation_id)
            # Sync request
            elif message.sync_request_msg.source_id == self.source_id:
                self._do_sync(send_msg, message.correlation_id,
                              message.sync_request_msg)
        # Do nothing for subscription ack or invalid message

    def _do_full_sync(self, send_msg, correlation_id):
        from vmware.nsx.monitoring.monitoring_provider_service_pb2 import \
            MonitoringSubClientResponseMsg
        from vmware.nsx.monitoring.monitoring_event_report_pb2 import \
            MonitoringFullSyncResponseMsg
        try:
            # Send event report
            try:
                self.service.event_reports_queue_lock.acquire()
                self.service.event_reports.append(
                    self.service.generate_full_event_report())
            finally:
                self.service.event_reports_queue_lock.release()
            # Respond to provider
            full_sync_response = MonitoringFullSyncResponseMsg()
            full_sync_response.result = True
            response_msg = MonitoringSubClientResponseMsg()
            response_msg.msg_type = 1
            response_msg.source_id = self.source_id
            response_msg.correlation_id = correlation_id
            response_msg.full_sync_response_msg.result = \
                full_sync_response.result
            send_msg(response_msg)
            lg.info("Successfully sent full sync response for %s",
                    self.source_id)
        except Exception:
            lg.warning("Full sync for %s failed with error reason: %s",
                       self.source_id, traceback.format_exc())

    def _do_sync(self, send_msg, correlation_id, sync_request):
        from vmware.nsx.monitoring.monitoring_provider_service_pb2 import \
            MonitoringSubClientResponseMsg
        from vmware.nsx.monitoring.monitoring_event_report_pb2 import \
            MonitoringSyncResponseMsg
        try:
            event_type_id = sync_request.event_type_id
            entity_id_msg = sync_request.entity_id
            entity_id_int = (entity_id_msg.left << 64) | entity_id_msg.right
            entity_id = uuid.UUID(int=entity_id_int)
            if event_type_id in self.service.event_type_statuses.keys() and \
                entity_id in self.service.event_type_statuses.get(event_type_id):
                status_tuple = self.service.event_type_statuses.get(
                    event_type_id).get(entity_id)
            else:
                lg.info("feature_id: {} event_type_id: {} has been deprecated "
                    "in this version, send status False to collector to clear "
                    "the stale alarms.",
                    self.service.feature_id, event_type_id)
                status_tuple = (False, {})
            # Respond to provider
            sync_response = MonitoringSyncResponseMsg()
            sync_response.status = status_tuple[0]
            sync_response.context = json.dumps(status_tuple[1])
            response_msg = MonitoringSubClientResponseMsg()
            response_msg.msg_type = 2
            response_msg.source_id = self.source_id
            response_msg.correlation_id = correlation_id
            response_msg.sync_response_msg.status = sync_response.status
            response_msg.sync_response_msg.context = sync_response.context
            send_msg(response_msg)
            lg.info("Successfully sent sync response for %s", self.source_id)
        except Exception:
            lg.warning("Sync for %s failed with error reason: %s",
                       self.source_id, traceback.format_exc())


class ScheduleManager(object):
    """Used to track when all Event type's next sample should be taken."""
    _NUM_WORKER_THREADS = 5  # Number of worker threads
    _WORKER_THREAD_INTERVAL = 1  # How often worker threads check for work
    _MAIN_THREAD_INTERVAL = 1  # How often main thread checks for work
    def __init__(self, event_source_service):
        self.started = False
        self.service = event_source_service
        # self.schedule_queue tracks when the next sample for all Event types
        # should be taken. It is a list of EventTypeSchedule instances sorted
        # by timestamp when next sample is taken, lowest timestamp first.
        self.schedule_queue = []
        # self.work_queue tracks all EventTypeSchedule instances that need
        # to be processed now
        self.work_queue = []
        self.schedule_lock = None
        self.sleep_method = None
        self.schedule_to_remove = []

    def start_scheduling(self):
        """Start ScheduleManager scheduling process it not already started."""
        if self.started:
            return
        # Initialize all threads needed to process events and the semaphore
        # used by the threads to synchronize access to shared data structures
        self.schedule_lock = BoundedSemaphore()
        self.sleep_method = gevent.sleep
        gevent.spawn(self._main_thread)
        for _ in range(self._NUM_WORKER_THREADS):
            gevent.spawn(self._worker_thread)
        # Initialize all the schedules
        schedule_details = self.service.get_event_type_schedule_details()
        for event_type_id, sampling_interval, poll_callback in schedule_details:
            self._add_new_schedule(
                event_type_id, sampling_interval, poll_callback)
        self.started = True

    def _add_schedule(self, schedule):
        """Add EventTypeSchedule to the schedule queue."""
        timestamp = schedule.get_next_sample_ts()
        num_schedules = len(self.schedule_queue)
        pos = 0
        added = False
        with EventLock(self.service, self.schedule_lock):
            try:
                while pos < num_schedules:
                    if timestamp < self.schedule_queue[pos].get_next_sample_ts():
                        self.schedule_queue.insert(pos, schedule)
                        added = True
                        break
                    pos += 1
                if not added:
                    self.schedule_queue.append(schedule)
            except Exception:
                lg.exception("Error adding a schedule to the schedule queue")

    def _add_new_schedule(self, event_type_id, sampling_interval,
                          poll_callback):
        """Add a new EventTypeSchedule to the schedule queue."""
        if not poll_callback:
            # For interrupt based events, check every 60 seconds to see if
            # there are samples reported by the feature interrupt
            sampling_interval = 60
        schedule = EventTypeSchedule(
            event_type_id, sampling_interval, poll_callback)
        self._add_schedule(schedule)

    def _update_schedule(self, event_type_id, sampling_interval,
                         poll_callback=None):
        """Update the sampling interval for an event type id."""
        self.schedule_to_remove.append(event_type_id)
        self._add_new_schedule(event_type_id, sampling_interval, poll_callback)

    def _main_thread(self):
        """Main scheduler thread that determines when an Event's interval is
           up and needs to be polled."""
        while True:
            self.sleep_method(self._MAIN_THREAD_INTERVAL)
            with EventLock(self.service, self.schedule_lock):
                try:
                    if sys.version_info.major == 2:
                        timestamp = time.time()
                    else:
                        timestamp = time.monotonic()
                    while self.schedule_queue:
                        if self.schedule_queue[0].get_next_sample_ts() > timestamp:
                            break
                        schedule = self.schedule_queue.pop(0)
                        self.work_queue.append(schedule)
                except Exception:
                    lg.exception(
                        "ScheduleManager._main_thread encountered an unexpected "
                        "error")

    def _worker_thread(self):
        """Worker thread that periodically looks in the work queue for Events
           that need to be polled and performs the poll."""
        while True:
            schedule = None
            with EventLock(self.service, self.schedule_lock):
                try:
                    if self.work_queue:
                        schedule = self.work_queue.pop(0)
                except Exception:
                    lg.exception(
                        "ScheduleManager._worker_thread encountered an unexpected "
                        "error processing the work queue")
            # If no events need to be processed, sleep until next interval
            if schedule is None:
                self.sleep_method(self._WORKER_THREAD_INTERVAL)
                continue
            # A schedule object needs to be processed
            event_type_id = schedule.get_event_type_id()
            try:
                # Check if the event type id is disabled
                is_disabled = self.service.event_type_config_sample_change[
                    event_type_id][0]
                if not is_disabled:
                    poll_callback = schedule.get_poll_callback()
                    if poll_callback is not None:
                        event_type_status_map = poll_callback([event_type_id])
                        if event_type_id in event_type_status_map:
                            self.service.process_new_sample(
                                {event_type_id:
                                     event_type_status_map[event_type_id]}, True)
                    else:
                        self.service.process_interrupt_based_samples(
                            event_type_id)
            except Exception:
                lg.exception("Error processing event poll")
            finally:
                # Skip an schedule if sampling interval changed because a new
                # schedule would have already been added
                if event_type_id in self.schedule_to_remove:
                    lg.info("Skip adding feature_id %d event_type_id %s to "
                            "schedule because sample interval has changed",
                            self.service.feature_id, event_type_id)
                    self.schedule_to_remove.remove(event_type_id)
                # Put the schedule back into the schedule queue if not disabled
                elif not is_disabled:
                    schedule.set_next_sample_ts()
                    self._add_schedule(schedule)
                else:
                    lg.info("Skip adding disabled feature_id %d event_type_id "
                            "%s to schedule", self.service.feature_id,
                            event_type_id)


class EventTypeSchedule(object):
    """
        Used to track when an Event type's next sample should be taken where:
            self.sampling_interval is the seconds between samples.
            self.next_sample_ts is timestamp of next sample in epoch seconds.
            self.poll_callback is the poll_callback of an is_poll_based event.
            self.is_poll_based is whether the event is a poll-based event.
        Note: this class models an Event not Event Intance.
    """
    def __init__(self, event_type_id, sampling_interval, poll_callback=None):
        self.event_type_id = event_type_id
        self.sampling_interval = sampling_interval
        self.poll_callback = poll_callback
        self.is_poll_based = poll_callback is not None
        self.set_next_sample_ts()

    def get_event_type_id(self):
        """Get ID of event type."""
        return self.event_type_id

    def get_sampling_interval(self):
        """Get sampling interval of event type."""
        return self.sampling_interval

    def get_poll_callback(self):
        """Get poll callback of event type, can return None."""
        return self.poll_callback

    def get_next_sample_ts(self):
        """Get timestamp of when next sample should be taken."""
        return self.next_sample_ts

    def set_next_sample_ts(self):
        """Update timestamp of when next sample should be taken."""
        if sys.version_info.major == 2:
            self.next_sample_ts = time.time() + self.sampling_interval
        else:
            self.next_sample_ts = time.monotonic() + self.sampling_interval

    def set_sampling_interval(self, sampling_interval):
        """Set sampling interval for the specifie event type."""
        self.sampling_interval = sampling_interval


class EventTypeHistory(object):
    """
        Used to track the Event status (True or False after considering
        hysteresis) of a given entity_id where:
            self.status is the status value after considering hysteresis.
            self.last_interrupt_status is the last reported sample status for a
                interrupt based event type.
            self.history is a cirular buffer of bools tracking samples reported.
            self.last_pos is the last position updated.
            self.sample_size is the number of sample for the event_type.
            self.sensitivity_percentage is the percentage (float) of samples
                that must be True for self.status to be True.
        Note: this class models an Event Instance not Event.
    """
    def __init__(self, is_poll_based, sample_size, sensitivity):
        """EventTypeHistory constructor, where sensitivity is 1-100"""
        self.is_poll_based = is_poll_based
        self.sample_size = sample_size
        self.history = [False] * sample_size
        self.last_pos = -1
        self.status = False
        self.last_interrupt_status = False
        self.set_sensitivity(sensitivity, update_status=False)

    def set_sensitivity(self, sensitivity, update_status=True):
        """Update sensitivity value.

            Arguments:
            - sensitivity: new sensitivity value in range 1-100 inclusive.

            Returns:
            - status after considering hysteresis of all samples based on
                the new sensitivity value.
        """
        self.sensitivity_percentage = float(sensitivity / 100)
        if update_status:
            self._update_status()
        return self.status

    def set_sample_size(self, sample_size, update_status=True):
        """Update sample_size value.

            Arguments:
            - sample_size: new sample_size value.

            Returns:
            - status after considering hysteresis of all samples based on
                the new sample_size value.
        """
        # Ignore invalid sample sizes
        if sample_size <= 0:
            return self.status
        # If new sample size is same as old, no processing needed
        if sample_size == self.sample_size:
            return self.status
        new_history = [False] * sample_size
        pos = self.last_pos
        if sample_size < self.sample_size:
            preserve_count = sample_size
        else:
            preserve_count = self.sample_size
        # Preserve the last reported samples in current history
        while preserve_count > 0:
            preserve_count -= 1
            new_history[preserve_count] = self.history[pos]
            pos -= 1
            if pos < 0:
                pos = self.sample_size - 1
        if sample_size < self.sample_size:
            self.last_pos = sample_size - 1
        else:
            self.last_pos = self.sample_size - 1
        self.sample_size = sample_size
        self.history = new_history
        if update_status:
            self._update_status()
        return self.status

    def set_last_interrupt_status(self, status):
        """Set last interrupt status for non poll based events."""
        self.last_interrupt_status = status

    def add_new_sample(self, sample_status=None):
        """Add new sample and return post hysteresis status after considering
           latest sample.

            Arguments:
            - sample_status: new boolean status to add

            Returns:
            - status after considering hysteresis of all samples including
                latest sample.
        """
        # Ignore add request if is_poll_based and no sample_status specified
        if self.is_poll_based and sample_status is None:
            lg.warning(
                "add_new_sample invoked without a sample_status for a "
                "is_poll_based eventy_type")
            return self.status
        elif not self.is_poll_based:
            sample_status = self.last_interrupt_status
        new_pos = self.last_pos + 1
        if new_pos >= self.sample_size:
            new_pos = 0
        self.history[new_pos] = sample_status
        self.last_pos = new_pos
        self._update_status()
        return self.status

    def _update_status(self):
        self.status = (float(self.history.count(True) / self.sample_size) >=
                       self.sensitivity_percentage)
        return self.status
