# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential
import os
import platform
import operator
import subprocess
import traceback
import sha.config
import socket
import re
import copy
import uuid
from sha.config import HOST_ID
from sha.contrib.metric.abc_base import Metric
from sha.contrib.metric.utils import run_command, timeout_cache
from sha.constants import APPLIANCE_PROXY, IP, ENTITY, UUID, SRC_NODE,\
    SRC_IP, DST_NODE, DST_IP, STATUS, UP, DOWN,\
    PACKET_TRANSMITTED, PACKET_RECEIVED, PACKET_LOSS_PERCENT, RTT_MIN, RTT_AVG,\
    RTT_MAX, RTT_MDEV, TOTAL_TIME, PERCENT, COUNT, DST_NODE_UUID, NODE_LATENCY_THRESHOLD

class UaNodeLatencyList(Metric):
    last_check = {}

    def __init__(self):
        self._logger = sha.config.get_logger(__name__)
        self.pat1 = re.compile(r'\d+')
        self.pat2 = re.compile(r'\d+\.\d+')

    @timeout_cache(2)
    def usage(self, *args, **kwargs):
        def get_nodes_by_config():
            path = r'/etc/vmware/nsx/appliance-info.xml'
            try:
                import xml.etree.ElementTree as ET
                root = ET.parse(path).getroot()
                cur_host = None
                cur_ip = None
                nodes = []
                for ele in root.iter(APPLIANCE_PROXY):
                    is_cur_node = False
                    ip = ele.find(IP).text
                    app_uuid = ele.find(UUID).text
                    if app_uuid == HOST_ID:
                        is_cur_node = True
                        cur_ip = ip
                    for entity in ele.iter(ENTITY):
                        uuid = entity.find(UUID).text
                        if is_cur_node:
                            cur_host = uuid
                        else:
                            nodes.append((uuid, ip))

                return cur_ip, cur_host, nodes
            except Exception as e:
                self._logger.error(
                    "could not read node-uuid from {}, error: {}".format(path, e))
                return None, []

        def inc_count(dst_node):
            data = UaNodeLatencyList.last_check.get(dst_node, {})
            count = data.get(COUNT, 0)
            count = count%5 + 1
            data[COUNT] = count
            return count

        def clear_count(dst_node):
            data = UaNodeLatencyList.last_check.get(dst_node, {})
            data[COUNT] = 0

        def get_ua_node_latency_list():
            cur_ip, cur_host, nodes = get_nodes_by_config()
            node_latency_list = []
            new_node = set()
            for node in nodes:
                dst_node = node[0]
                dst_ip = node[1]
                status = {SRC_NODE: cur_host,
                          SRC_IP: cur_ip,
                          DST_NODE: dst_node,
                          DST_IP: dst_ip}

                if dst_node not in UaNodeLatencyList.last_check:
                    data = copy.deepcopy(status)
                    data[DST_NODE_UUID] = uuid.UUID(dst_node)
                    data[COUNT] = 0
                    UaNodeLatencyList.last_check[dst_node] = data

                new_node.add(dst_node)

                ping_cmd = ["ping", "-c", "4", "-q", dst_ip]
                try:
                    output = run_command(cmd=ping_cmd, shell=False, timeout=10)
                except:
                    output = None

                if not output:
                    status[PACKET_LOSS_PERCENT] = '100%'
                    status[STATUS] = DOWN
                    inc_count(dst_node)
                else:
                    data = output.strip().split(os.linesep)[-2:]
                    packet_data = self.pat1.findall(data[0])
                    status[PACKET_TRANSMITTED] = int(packet_data[0])
                    status[PACKET_RECEIVED] = int(packet_data[1])
                    status[PACKET_LOSS_PERCENT] = packet_data[2]+PERCENT
                    status[TOTAL_TIME] = int(packet_data[3])

                    latency_data = [float(i) for i in self.pat2.findall(data[1])]
                    status[RTT_MIN] = latency_data[0]
                    status[RTT_AVG] = latency_data[1]
                    status[RTT_MAX] = latency_data[2]
                    status[RTT_MDEV] = latency_data[3]
                    status[STATUS] = UP

                    if status[RTT_AVG] >= NODE_LATENCY_THRESHOLD:
                        inc_count(dst_node)
                    else:
                        clear_count(dst_node)

                node_latency_list.append(status)

            removed = set(UaNodeLatencyList.last_check) - new_node
            alarm_status = copy.deepcopy(UaNodeLatencyList.last_check)
            for node in removed:
                data = alarm_status.get(node)
                data[COUNT] = 0
                UaNodeLatencyList.last_check.pop(node, None)

            node_latency_list.append(alarm_status)
            self._logger.debug("node status: {}".format(node_latency_list))
            return node_latency_list
        return get_ua_node_latency_list()