# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential

import os
import struct
import traceback

import gzip
import shutil

import sha.preconfig
from sha.core.channel.provider.base import BaseProvider
from sha.constants import DATA
from sha import config

FILE_ROTATION_MAX_BYTES = 1024 * 1024
FILE_ROTATION_BACKUP_COUNT = 10

class FiledumpProvider(BaseProvider):
    """The File dump Provider"""

    def __init__(self, filename, dump_callback=None, init_callback=None,
                 max_bytes=FILE_ROTATION_MAX_BYTES, backup_count=FILE_ROTATION_BACKUP_COUNT):
        self._name = "FiledumpProvider"
        self._logger = sha.preconfig.get_sub_logger(__name__, s2comp = "FiledumpProvider")
        self._filename = os.path.join(config.DUMP_FILE_PATH, filename)
        self._dump_callback = dump_callback
        self._init_callback = init_callback
        self._rotator = FileRotatingHandler(self._filename, init_callback, max_bytes, backup_count)
        self._logger.info("Provider to dump data to file {} by {} and {}"
                          .format(self._filename, self._dump_callback, self._init_callback))

    def sync(self, msg):
        try:
            self._logger.debug("msg:{}".format(msg))
            if self._dump_callback:
                data = self._dump_callback(msg[DATA])
            else:
                data = msg[DATA]
            self._rotator.dump(data)

        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("{} failed in dump data to file:{}, Exception:{}, trace:{}"
                                 .format(self._name, msg, e, trace))

    def close(self):
        self._rotator._close()

    @property
    def name(self):
        return self._name

class BaseRotatingHandler():
    """Base rotating handler"""

    def __init__(self, filename, init_callback):
        self._logger = sha.preconfig.get_sub_logger(__name__, s2comp = self._name)
        self._base_filename = filename
        self._init_callback = init_callback
        self._stream = None

    def _open(self):
        if self._stream:
            self._logger.warning("File has been opened!")
            return

        # Check whether file existing
        new_file = False
        if not os.path.exists(self._base_filename):
            new_file = True
            self._logger.info("No {} file, will create one".format(self._base_filename))
        self._stream = open(self._base_filename, "ab")
        if new_file and self._init_callback:
            self._stream.write(self._init_callback())

    def _close(self):
        if self._stream:
            self._stream.close()
            self._stream = None

    def _write(self, data):
        # Write data to file
        self._stream.write(data)
        self._stream.flush()

    def dump(self, data):
        """
        Dump the data to the file, catering for rollover as described
        in doRollover().
        """
        self._logger.info("Dumpping data: {}".format(data))
        try:
            if self._shouldRollover(data):
                self._doRollover()
            self._write(data)
            self._logger.debug("Dummpped data: {}".format(data))
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to dump data to file:{}, Exception:{}, trace:{}"
                                 .format(data, e, trace))

    def _rotate(self, sfn, dfn):
        with open(sfn, 'rb') as f_in:
            with gzip.open(dfn, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
        os.remove(sfn)
        self._logger.info("Compress {} into {}".format(sfn, dfn))


class FileRotatingHandler(BaseRotatingHandler):
    """The file rotating handler"""

    def __init__(self, filename, init_callback, max_bytes, backup_count):
        self._name = "FileRotator"
        BaseRotatingHandler.__init__(self, filename, init_callback)
        self._max_bytes = max_bytes
        self._backup_count = backup_count
        # if base_file_name with extension, the backup name should be pure_filename.x.extension.gz
        if '.' in filename:
            self._backup_base = filename.rsplit('.')[0]
            self._backup_ext = '.' + filename.rsplit('.')[1]
        else:
            self._backup_base = filename
            self._backup_ext = ""
        self._logger.info("Create file rotator: basefilename: {}, backup filename: {}.x{}.gz,  max_bytes: {}, backup_count: {}"
                           .format(self._base_filename, self._backup_base, self._backup_ext,
                                   self._max_bytes, self._backup_count))

    def _shouldRollover(self, data):
        """
        Determine if rollover should occur.
        Basically, see if the supplied data would cause the file to exceed
        the size limit we have.
        """
        if self._stream is None:
            self._open()
        elif not os.path.exists(self._base_filename):
            self._logger.warning("Opened file {} has gone away.".format(self._base_filename))
            self._close()
            self._open()

        if self._max_bytes > 0:
            self._stream.seek(0, 2)
            if self._stream.tell() + len(data) >= self._max_bytes:
                self._logger.info("Current file size {} + data size {} reach the rotation limit: {}"
                                  .format(self._stream.tell(), len(data), self._max_bytes))
                return True
        return False

    def _doRollover(self):
        self._close()

        # Rollover backup files
        for i in range(self._backup_count - 2, -1, -1):
            sfn = "%s.%d%s.gz" % (self._backup_base, i, self._backup_ext)
            if not os.path.exists(sfn):
                continue
            dfn = "%s.%d%s.gz" % (self._backup_base, i + 1, self._backup_ext)
            if os.path.exists(dfn):
                os.remove(dfn)
            os.rename(sfn, dfn)
            self._logger.info("Rename {} to {}".format(sfn, dfn))

        # Rotate the base file
        dfn = "%s.%d%s.gz" % (self._backup_base, 0, self._backup_ext)
        self._rotate(self._base_filename, dfn)
        self._logger.info("Rotate {} to {}".format(self._base_filename, dfn))

        self._open()

    def close(self):
        self._close()

    @property
    def name(self):
       return self._name

# workflow trace span dumping callback
def trace_span_init_callback():
    magic = bytes([0x6e, 0x3c, 0xb9, 0x83, 0x5b, 0xdd, 0x4f, 0x3b, 0x91, 0x80, 0x55, 0x89, 0xe3, 0xd4, 0xf5, 0xda])
    fileHdr = struct.pack('16sBBH', magic, 1, 0, 0)
    return fileHdr

def trace_span_dump_callback(data):
    dataHdr = struct.pack('HBB', len(data), 0, 0)
    return dataHdr + data

# metric dumping callback for test
def metric_init_callback():
    str = "Begin of a new file!\n"
    return str.encode()

def metric_dump_callback(data):
    import time
    str = '{}: {}\n'.format(time.time(),data)
    return str.encode()
