# Copyright (C) 2019 VMware, Inc. All rights reserved.
# -- VMware Confidential
import functools
import heapq
import json

from sha.constants import ENABLE, PROFILE_NAME
import sha.config


class ProfileCollection:
    """
    The collection of profiles of a plugin.
    Each profile will be sorted by priority. Highest priority (lowest value) profile will always be the 1st item of list.
    """
    _logger = sha.config.get_logger(__name__)

    def __init__(self):
        # profile list sorted by priority of profile
        self._sorted_profiles = []

    def add_or_update(self, profile_msg_id, priority, name, enabled, content):
        exist_p = self._find(profile_msg_id)
        if exist_p is None:
            new_p = Profile(profile_msg_id, priority, name, enabled, content)
            heapq.heappush(self._sorted_profiles, new_p)
            ProfileCollection._logger.info("Add profile succeed, profile_msg_id:{}, all profiles in "
                                           "ProfileCollection: {}".format(profile_msg_id, self._sorted_profiles))
            return

        exist_p.update_property(priority, name, enabled, content)
        heapq.heapify(self._sorted_profiles)
        ProfileCollection._logger.info("Update profile succeed, profile_msg_id:{}, all profiles in "
                                       "ProfileCollection: {}".format(profile_msg_id, self._sorted_profiles))

    def get(self):
        """
        get the highest priority (smallest value) profile
        """
        first_p = self._sorted_profiles[0]
        res = {}
        res.update(first_p.content)
        res[PROFILE_NAME] = first_p.name
        res[ENABLE] = first_p.enabled
        ProfileCollection._logger.debug("Got profile: {} from ProfileCollection:{}".format(res, self._sorted_profiles))
        return res

    def delete(self, profile_msg_id):
        """
        delete profile by id
        """
        for p in self._sorted_profiles:
            if p.profile_msg_id == profile_msg_id:
                self._sorted_profiles.remove(p)
                heapq.heapify(self._sorted_profiles)
                break
        ProfileCollection._logger.info("After delete profile_msg_id:{}, ProfileCollection: {}"
                                       .format(profile_msg_id, self._sorted_profiles))

    def _find(self, profile_msg_id):
        """
        find profile by id
        """
        for p in self._sorted_profiles:
            if p.profile_msg_id == profile_msg_id:
                return p
        return None

    def is_empty(self):
        return len(self._sorted_profiles) == 0


@functools.total_ordering
class Profile:
    def __init__(self, profile_msg_id, priority, name, enabled, content):
        self._profile_msg_id = profile_msg_id
        self._priority = priority
        self._name = name
        self._enabled = enabled
        self._content = content

    def __repr__(self):
        return "Profile[profile_msg_id:{}, name:{}, priority:{}, enabled:{}, content:{}"\
            .format(self._profile_msg_id, self.name, self.priority,self.enabled, self.content)

    def update_property(self, priority, name, enabled, content):
        self._priority = priority
        self._name = name
        self._enabled = enabled
        self._content = content

    @property
    def profile_msg_id(self):
        return self._profile_msg_id

    @property
    def name(self):
        return self._name

    @property
    def priority(self):
        return self._priority

    @property
    def enabled(self):
        return self._enabled

    @property
    def content(self):
        return self._content

    def __eq__(self, other):
        return self.priority == other.priority

    def __lt__(self, other):
        return self.priority < other.priority
